cimport cython

from cpython.object cimport PyObject
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem


DEF NUM_SCORES = 9


@cython.freelist(100000)
cdef class ScoreSet(object):
    def __init__(self, glycopeptide_score=0., peptide_score=0., glycan_score=0., glycan_coverage=0.,
                 stub_glycopeptide_intensity_utilization=0., oxonium_ion_intensity_utilization=0.,
                 n_stub_glycopeptide_matches=0, peptide_coverage=0.0, total_signal_utilization=0.0):
        self.glycopeptide_score = glycopeptide_score
        self.peptide_score = peptide_score
        self.glycan_score = glycan_score
        self.glycan_coverage = glycan_coverage

        self.stub_glycopeptide_intensity_utilization = stub_glycopeptide_intensity_utilization
        self.oxonium_ion_intensity_utilization = oxonium_ion_intensity_utilization
        self.n_stub_glycopeptide_matches = n_stub_glycopeptide_matches

        self.peptide_coverage = peptide_coverage
        self.total_signal_utilization = total_signal_utilization

    def to_dict(self):
        store = {}
        for name, value in zip(self.field_names(), self.values()):
            store[name] = value
        return store

    cpdef bytearray pack(self):
        cdef:
            float[NUM_SCORES] data
        data[0] = self.glycopeptide_score
        data[1] = self.peptide_score
        data[2] = self.glycan_score
        data[3] = self.glycan_coverage
        data[4] = self.stub_glycopeptide_intensity_utilization
        data[5] = self.oxonium_ion_intensity_utilization
        data[6] = self.n_stub_glycopeptide_matches
        data[7] = self.peptide_coverage
        data[8] = self.total_signal_utilization
        return ((<char*>data)[:sizeof(float) * NUM_SCORES])

    @staticmethod
    def unpack(bytearray data):
        cdef:
            float* buff
            char* temp
            int n_stub_glycopeptide_matches
            float stub_utilization
            float oxonium_utilization
            float peptide_coverage
            float total_signal_utilization
        temp = data
        buff = <float*>(temp)
        stub_utilization = buff[4]
        oxonium_utilization = buff[5]
        n_stub_glycopeptide_matches = <int>buff[6]
        peptide_coverage = buff[7]
        total_signal_utilization = buff[8]
        return ScoreSet._create(buff[0], buff[1], buff[2], buff[3], stub_utilization,
                                oxonium_utilization, n_stub_glycopeptide_matches, peptide_coverage,
                                total_signal_utilization)

    @staticmethod
    cdef ScoreSet _create(float glycopeptide_score, float peptide_score, float glycan_score, float glycan_coverage,
                          float stub_glycopeptide_intensity_utilization, float oxonium_ion_intensity_utilization,
                          int n_stub_glycopeptide_matches, float peptide_coverage, float total_signal_utilization):
        cdef:
            ScoreSet self
        self = ScoreSet.__new__(ScoreSet)
        self.glycopeptide_score = glycopeptide_score
        self.peptide_score = peptide_score
        self.glycan_score = glycan_score
        self.glycan_coverage = glycan_coverage
        self.stub_glycopeptide_intensity_utilization = stub_glycopeptide_intensity_utilization
        self.oxonium_ion_intensity_utilization = oxonium_ion_intensity_utilization
        self.n_stub_glycopeptide_matches = n_stub_glycopeptide_matches
        self.peptide_coverage = peptide_coverage
        self.total_signal_utilization = total_signal_utilization
        return self

    def __eq__(self, other):
        return self._eq(other)

    def __ne__(self, other):
        return not self._eq(other)

    def __lt__(self, other):
        return self._lt(other)

    def __gt__(self, other):
        return self._gt(other)

    def __repr__(self):
        template = (
            "{self.__class__.__name__}({self.glycopeptide_score}, {self.peptide_score},"
            " {self.glycan_score}, {self.glycan_coverage}, {self.stub_glycopeptide_intensity_utilization},"
            " {self.oxonium_ion_intensity_utilization}, {self.n_stub_glycopeptide_matches}, {self.peptide_coverage},"
            " {self.total_signal_utilization})")
        return template.format(self=self)

    def __len__(self):
        return 7

    def __getitem__(self, int i):
        if i == 0:
            return self.glycopeptide_score
        elif i == 1:
            return self.peptide_score
        elif i == 2:
            return self.glycan_score
        elif i == 3:
            return self.glycan_coverage
        elif i == 4:
            return self.stub_glycopeptide_intensity_utilization
        elif i == 5:
            return self.oxonium_ion_intensity_utilization
        elif i == 6:
            return self.n_stub_glycopeptide_matches
        elif i == 7:
            return self.peptide_coverage
        elif i == 8:
            return self.total_signal_utilization
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.glycopeptide_score
        yield self.peptide_score
        yield self.glycan_score
        yield self.glycan_coverage
        yield self.stub_glycopeptide_intensity_utilization
        yield self.oxonium_ion_intensity_utilization
        yield self.n_stub_glycopeptide_matches
        yield self.peptide_coverage
        yield self.total_signal_utilization

    def __reduce__(self):
        return self.__class__, (self.glycopeptide_score, self.peptide_score,
                                self.glycan_score, self.glycan_coverage, self.stub_glycopeptide_intensity_utilization,
                                self.oxonium_ion_intensity_utilization, self.n_stub_glycopeptide_matches,
                                self.peptide_coverage, self.total_signal_utilization)

    @classmethod
    def from_spectrum_matcher(cls, match):
        stub_utilization, n_stub_glycopeptide_matches, _, total_signal_utilization = match.count_peptide_Y_ion_utilization()
        oxonium_utilization = match.oxonium_ion_utilization()
        return cls(match.score,
            match.peptide_score(),
            match.glycan_score(),
            match.glycan_coverage(),
            stub_utilization,
            oxonium_utilization,
            n_stub_glycopeptide_matches,
            match.peptide_coverage(),
            total_signal_utilization,
        )

    @classmethod
    def field_names(cls):
        return [
            "total_score",
            "peptide_score",
            "glycan_score",
            "glycan_coverage",
            "stub_glycopeptide_intensity_utilization",
            "oxonium_ion_intensity_utilization",
            "n_stub_glycopeptide_matches",
            "peptide_coverage",
            "total_signal_utilization",
        ]

    cpdef list values(self):
        return [
            self.glycopeptide_score,
            self.peptide_score,
            self.glycan_score,
            self.glycan_coverage,
            self.stub_glycopeptide_intensity_utilization,
            self.oxonium_ion_intensity_utilization,
            self.n_stub_glycopeptide_matches,
            self.peptide_coverage,
            self.total_signal_utilization,
        ]

    cpdef bint _eq(self, ScoreSet other):
        if abs(self.glycopeptide_score - other.glycopeptide_score) > 1e-3:
            return False
        if abs(self.peptide_score - other.peptide_score) > 1e-3:
            return False
        if abs(self.glycan_score - other.glycan_score) > 1e-3:
            return False
        if abs(self.glycan_coverage - other.glycan_coverage) > 1e-3:
            return False
        return True

    cpdef bint _lt(self, ScoreSet other):
        if self.glycopeptide_score < other.glycopeptide_score:
            return True
        elif abs(self.glycopeptide_score - other.glycopeptide_score) > 1e-3:
            return False

        if self.peptide_score < other.peptide_score:
            return True
        elif abs(self.peptide_score - other.peptide_score) > 1e-3:
            return False

        if self.glycan_score < other.glycan_score:
            return True
        elif abs(self.glycan_score - other.glycan_score) > 1e-3:
            return False

        if self.glycan_coverage < other.glycan_coverage:
            return True
        return False

    cpdef bint _gt(self, ScoreSet other):
        if self.glycopeptide_score > other.glycopeptide_score:
            return True
        elif abs(self.glycopeptide_score - other.glycopeptide_score) > 1e-3:
            return False

        if self.peptide_score > other.peptide_score:
            return True
        elif abs(self.peptide_score - other.peptide_score) > 1e-3:
            return False

        if self.glycan_score > other.glycan_score:
            return True
        elif abs(self.glycan_score - other.glycan_score) > 1e-3:
            return False

        if self.glycan_coverage > other.glycan_coverage:
            return True
        return False


@cython.freelist(100000)
@cython.final
cdef class FDRSet(object):

    def __init__(self, total_q_value, peptide_q_value, glycan_q_value, glycopeptide_q_value):
        self.total_q_value = total_q_value
        self.peptide_q_value = peptide_q_value
        self.glycan_q_value = glycan_q_value
        self.glycopeptide_q_value = glycopeptide_q_value

    def __eq__(self, other):
        return self._eq(other)

    def __ne__(self, other):
        return not self._eq(other)

    def __lt__(self, other):
        return self._lt(other)

    def __gt__(self, other):
        return self._gt(other)

    def __repr__(self):
        template = (
            "{self.__class__.__name__}({self.total_q_value}, {self.peptide_q_value}, "
            "{self.glycan_q_value}, {self.glycopeptide_q_value})")
        return template.format(self=self)

    cpdef bint _eq(self, FDRSet other):
        if other is None:
            return False
        if self is None:
            return False
        if abs(self.total_q_value - other.total_q_value) > 1e-3:
            return False
        if abs(self.peptide_q_value - other.peptide_q_value) > 1e-3:
            return False
        if abs(self.glycan_q_value - other.glycan_q_value) > 1e-3:
            return False
        if abs(self.glycopeptide_q_value - other.glycopeptide_q_value) > 1e-3:
            return False
        return True

    cpdef bint _lt(self, FDRSet other):
        if self.total_q_value < other.total_q_value:
            return True
        elif abs(self.total_q_value - other.total_q_value) > 1e-3:
            return False

        if self.peptide_q_value < other.peptide_q_value:
            return True
        elif abs(self.peptide_q_value - other.peptide_q_value) > 1e-3:
            return False

        if self.glycan_q_value < other.glycan_q_value:
            return True
        elif abs(self.glycan_q_value - other.glycan_q_value) > 1e-3:
            return False

        if self.glycopeptide_q_value < other.glycopeptide_q_value:
            return True
        return False

    cpdef bint _gt(self, FDRSet other):
        if self.total_q_value > other.total_q_value:
            return True
        elif abs(self.total_q_value - other.total_q_value) > 1e-3:
            return False

        if self.peptide_q_value > other.peptide_q_value:
            return True
        elif abs(self.peptide_q_value - other.peptide_q_value) > 1e-3:
            return False

        if self.glycan_q_value > other.glycan_q_value:
            return True
        elif abs(self.glycan_q_value - other.glycan_q_value) > 1e-3:
            return False

        if self.glycopeptide_q_value > other.glycopeptide_q_value:
            return True
        return False

    @staticmethod
    cdef FDRSet _create(double total_q_value, double peptide_q_value, double glycan_q_value, double glycopeptide_q_value):
        cdef:
            FDRSet self
        self = FDRSet.__new__(FDRSet)
        self.total_q_value = total_q_value
        self.peptide_q_value = peptide_q_value
        self.glycan_q_value = glycan_q_value
        self.glycopeptide_q_value = glycopeptide_q_value
        return self

    cpdef bytearray pack(self):
        cdef:
            double[4] data
        data[0] = self.total_q_value
        data[1] = self.peptide_q_value
        data[2] = self.glycan_q_value
        data[3] = self.glycopeptide_q_value
        return ((<char*>data)[:sizeof(double) * 4])

    @staticmethod
    def unpack(bytearray data):
        cdef:
            double* buff
            char* temp
        temp = data
        buff = <double*>(temp)
        return FDRSet._create(buff[0], buff[1], buff[2], buff[3])

    @classmethod
    def default(cls):
        return FDRSet._create(1.0, 1.0, 1.0, 1.0)

    def __len__(self):
        return 4

    def __getitem__(self, int i):
        if i == 0:
            return self.total_q_value
        elif i == 1:
            return self.peptide_q_value
        elif i == 2:
            return self.glycan_q_value
        elif i == 3:
            return self.glycopeptide_q_value
        else:
            raise IndexError(i)

    def __reduce__(self):
        return self.__class__, (self.total_q_value, self.peptide_q_value,
                                self.glycan_q_value, self.glycopeptide_q_value)

    def to_dict(self):
        store = {
            "total_q_value": self.total_q_value,
            "peptide_q_value": self.peptide_q_value,
            "glycan_q_value": self.glycan_q_value,
            "glycopeptide_q_value": self.glycopeptide_q_value
        }
        return store


@cython.final
cdef class PeakLabelMap(object):

    def __init__(self, name_to_peaks=None):
        if name_to_peaks is None:
            name_to_peaks = dict()
        self.name_to_peaks = name_to_peaks

    @staticmethod
    cdef PeakLabelMap _create():
        cdef PeakLabelMap self = PeakLabelMap.__new__(PeakLabelMap)
        self.name_to_peaks = {}
        return self

    cdef inline void add(self, str name, DeconvolutedPeak peak):
        cdef:
            list acc
        PyDict_SetItem(self.name_to_peaks, name, peak)

    cdef inline DeconvolutedPeak get(self, str name, bint* found):
        cdef:
            PyObject* temp
            DeconvolutedPeak result

        temp = PyDict_GetItem(self.name_to_peaks, name)
        if temp == NULL:
            found[0] = False
            return None
        else:
            result = <DeconvolutedPeak>temp
            found[0] = True
            return result


from collections import Counter

from itertools import chain
import math

from .base import GlycopeptideSpectrumMatcherBase
from glycresoft.structure import FragmentMatchMap

from glycopeptidepy.structure.fragment import IonSeries


def flatten(iterable):
    return [a for b in iterable for a in b]


class FrequencyCounter(object):
    def __init__(self):
        self.observed_pairs = Counter()
        self.observed_n_term = Counter()
        self.observed_c_term = Counter()
        self.observed_sequences = Counter()

        self.possible_pairs = Counter()
        self.possible_n_term = Counter()
        self.possible_c_term = Counter()

        self.series = tuple("by")

    def add_sequence(self, glycopeptide):
        self.observed_sequences[glycopeptide] += 1

    def add_fragment(self, fragment):
        n_term, c_term = fragment.flanking_amino_acids
        self.observed_pairs[n_term, c_term] += 1
        self.observed_n_term[n_term] += 1
        self.observed_c_term[c_term] += 1

    def total_possible_outcomes(self):
        possible_pairs = Counter()
        possible_n_term = Counter()
        possible_c_term = Counter()
        for seq, count in self.observed_sequences.items():
            for series in self.series:
                for fragment in flatten(seq.get_fragments(series)):
                    n_term, c_term = fragment.flanking_amino_acids
                    possible_pairs[n_term, c_term] += count
                    possible_n_term[n_term] += count
                    possible_c_term[c_term] += count

        self.possible_pairs = possible_pairs
        self.possible_n_term = possible_n_term
        self.possible_c_term = possible_c_term

    def process_match(self, spectrum_match):
        for case in spectrum_match.solution_map:
            if case.series == 'b' or case.series == 'y':
                self.add_fragment(case)
        self.add_sequence(spectrum_match.target)

    def n_term_probability(self, residue=None):
        if residue is not None:
            return self.observed_n_term[residue] / float(self.possible_n_term[residue])
        else:
            return {r: self.n_term_probability(r) for r in self.observed_n_term}

    def c_term_probability(self, residue=None):
        if residue is not None:
            return self.observed_c_term[residue] / float(self.possible_c_term[residue])
        else:
            return {r: self.c_term_probability(r) for r in self.observed_c_term}

    def pair_probability(self, pair=None):
        if pair is not None:
            pair = tuple(pair)
            return self.observed_pairs[pair] / float(self.possible_pairs[pair])
        else:
            return {r: self.pair_probability(r) for r in self.observed_pairs}

    def residue_probability(self, residue=None):
        if residue is not None:
            return (self.observed_n_term[residue] + self.observed_c_term[residue]) / float(
                self.possible_n_term[residue] + self.possible_c_term[residue])
        else:
            return {r: self.residue_probability(r) for r in (set(self.observed_c_term) | set(self.observed_n_term))}

    @classmethod
    def fit(cls, spectrum_match_set):
        inst = cls()
        for match in spectrum_match_set:
            inst.process_match(match)
        inst.total_possible_outcomes()
        return inst


class FrequencyScorer(GlycopeptideSpectrumMatcherBase):
    def __init__(self, scan, sequence, model=None, mass_shift=None):
        super(FrequencyScorer, self).__init__(scan, sequence, mass_shift)
        self._score = None
        self.solution_map = FragmentMatchMap()
        self.glycosylated_b_ion_count = 0
        self.glycosylated_y_ion_count = 0
        self.model = model

    def match(self, error_tolerance=2e-5):
        solution_map = FragmentMatchMap()
        spectrum = self.spectrum

        n_glycosylated_b_ions = 0
        for frags in self.target.get_fragments('b'):
            glycosylated_position = False
            for frag in frags:
                glycosylated_position |= frag.is_glycosylated
                peak = spectrum.has_peak(frag.mass, error_tolerance)
                if peak:
                    solution_map.add(peak, frag)
            if glycosylated_position:
                n_glycosylated_b_ions += 1

        n_glycosylated_y_ions = 0
        for frags in self.target.get_fragments('y'):
            glycosylated_position = False
            for frag in frags:
                glycosylated_position |= frag.is_glycosylated
                peak = spectrum.has_peak(frag.mass, error_tolerance)
                if peak:
                    solution_map.add(peak, frag)
            if glycosylated_position:
                n_glycosylated_y_ions += 1
        for frag in self.target.stub_fragments(extended=True):
            peak = spectrum.has_peak(frag.mass, error_tolerance)
            if peak:
                solution_map.add(peak, frag)

        self.glycosylated_b_ion_count = n_glycosylated_b_ions
        self.glycosylated_y_ion_count = n_glycosylated_y_ions
        self.solution_map = solution_map
        return solution_map

    def _compute_total(self):
        total = 0.
        for frags in chain(self.target.get_fragments('b'), self.target.get_fragments('y')):
            for frag in frags:
                n_term, c_term = frag.flanking_amino_acids
                score = self.frequency_counter.n_term_probability(
                    n_term) * self.frequency_counter.c_term_probability(c_term)
                total += score * 0.5
        return total

    def _score_backbone(self):
        total = self._compute_total()
        observed = 0.0
        track_site = set()
        for frag in self.solution_map.fragments():
            if (frag.series == 'b') or (frag.series == 'y'):
                position = frag.position
                n_term, c_term = frag.flanking_amino_acids
                score = self.model.n_term_probability(
                    n_term) * self.model.c_term_probability(c_term)
                weight = 0.6 if position not in track_site else 0.4
                track_site.add(position)
                observed += score * weight
        return observed / total



import unittest

from glypy.composition import composition_transform
from glypy.structure import glycan_composition

from glycresoft import symbolic_expression


class SymbolicExpressionTest(unittest.TestCase):
    def test_normalize_glycan_composition(self):
        base = glycan_composition.GlycanComposition.parse("{Hex:6; HexNAc:5; Neu5Ac:3}")
        deriv = composition_transform.derivatize(
            glycan_composition.GlycanComposition.parse(
                "{Hex:6; HexNAc:5; Neu5Ac:3}"), "methyl")

        normd_symbol = symbolic_expression.GlycanSymbolContext(deriv)
        normd_composition = glycan_composition.GlycanComposition.parse(normd_symbol.serialize())

        self.assertEqual(base, normd_composition)

    def test_complex_expression(self):
        ex = symbolic_expression.parse_expression("X + 1 + abs(-2Z) * 5")
        ctx = symbolic_expression.SymbolContext({"X": 5, "Z": 3})
        expected = 36
        self.assertEqual(ex.evaluate(ctx), expected)

    def test_nested_sub_expression(self):
        expr = symbolic_expression.parse_expression("(2(x + 5)) + 5")
        assert expr.evaluate({'x': 0}) == 15
        assert expr.evaluate({"x": 5}) == 25

    def test_nested_simplify(self):
        expr = symbolic_expression.parse_expression("(((x + 5)))")
        assert isinstance(expr.expr, symbolic_expression.ExpressionNode)
        expr = symbolic_expression.parse_expression("((((2(x + 5)))))")
        assert expr.coefficient == 2
        assert isinstance(expr.expr, symbolic_expression.ExpressionNode)


if __name__ == '__main__':
    unittest.main()



# cython: embedsignature=True

import operator

cimport cython
from cpython.list cimport PyList_Append, PyList_Size, PyList_GetItem


@cython.freelist(1000000)
cdef class MassObject(object):

    def __init__(self, obj, mass):
        self.obj = obj
        self.mass = mass

    def __repr__(self):
        return "MassObject(%r, %r)" % (self.obj, self.mass)

    def __lt__(self, other):
        return self.mass < other.mass

    def __gt__(self, other):
        return self.mass > other.mass

    def __eq__(self, other):
        return abs(self.mass - other.mass) < 1e-3

    def __ne__(self, other):
        return abs(self.mass - other.mass) >= 1e-3


def identity(x):
    return x


cdef class NeutralMassDatabaseImpl(object):

    def __init__(self, structures, mass_getter=operator.attrgetter("calculated_mass"), sort=True):
        self.mass_getter = mass_getter
        self.structures = self._prepare(structures, sort=sort)

    cdef list _prepare(self, list structures, bint sort=True):
        cdef:
            size_t i, n
            list translated
            MassObject mo
            object o

        translated = []
        n = PyList_Size(structures)
        for i in range(n):
            o = structures[i]
            mo = MassObject(o, self.mass_getter(o))
            translated.append(mo)
        if sort:
            translated.sort()
        return translated

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cpdef size_t search_binary(self, double mass, double error_tolerance=1e-6):
        """Search within :attr:`structures` for the index of a structure
        with a mass nearest to `mass`, within `error_tolerance`

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The approximate error tolerance to accept

        Returns
        -------
        int
            The index of the structure with the nearest mass
        """
        cdef:
            ssize_t lo, hi, n, mid, best_index, i
            double err, best_error
            MassObject x
            list array
        array = self.structures
        lo = 0
        n = hi = PyList_Size(array)
        while hi != lo:
            mid = (hi + lo) // 2
            x = <MassObject>PyList_GetItem(array, mid)
            err = (x.mass - mass) / mass
            if abs(err) <= error_tolerance:
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = <MassObject>PyList_GetItem(array, i)
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1

                i = mid + 1
                while i < n:
                    x = <MassObject>PyList_GetItem(array, i)
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif (hi - lo) == 1:
                best_index = mid
                best_error = err
                i = mid - 1
                while i >= 0:
                    x = <MassObject>PyList_GetItem(array, i)
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1

                i = mid + 1
                while i < n:
                    x = <MassObject>PyList_GetItem(array, i)
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif err > 0:
                hi = mid
            elif err < 0:
                lo = mid
        return 0

    def __len__(self):
        return self.get_size()

    def __iter__(self):
        for obj in self.structures:
            yield obj.obj

    def __getitem__(self, i):
        if isinstance(i, slice):
            return [mo.obj for mo in self.structures[i]]

        if i < 0:
            i = self.get_size() + i
        return self.get_item(i)

    def __reduce__(self):
        return self.__class__, ([], identity, False), self.__getstate__()

    @property
    def lowest_mass(self):
        return self.structures[0].mass

    @property
    def highest_mass(self):
        return self.structures[-1].mass

    cdef size_t get_size(self):
        return PyList_Size(self.structures)

    cdef object get_item(self, size_t i):
        cdef:
            MassObject mo
        mo = <MassObject>PyList_GetItem(self.structures, i)
        return mo.obj

    cpdef list search_mass(self, double mass, double error_tolerance=0.1):
        """Search for the set of all items in :attr:`structures` within `error_tolerance` Da
        of the queried `mass`.

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The range of mass errors (in Daltons) to allow

        Returns
        -------
        list
            The list of instances which meet the criterion
        """
        cdef:
            double lo_mass, hi_mass
            size_t lo, hi, i
            list result
            MassObject mo

        if self.get_size() == 0:
            return []
        lo_mass = mass - error_tolerance
        hi_mass = mass + error_tolerance
        lo = self.search_binary(lo_mass)
        hi = self.search_binary(hi_mass) + 1

        result = []
        for i in range(lo, hi):
            mo = <MassObject>PyList_GetItem(self.structures, i)
            if lo_mass <= mo.mass <= hi_mass:
                result.append(mo.obj)
        return result

    cpdef object search_between(self, double lower, double higher):
        cdef:
            size_t lo, hi, i
            list result
            MassObject mo
            NeutralMassDatabaseImpl inst

        if self.get_size() == 0:
            return []
        lo = self.search_binary(lower)
        hi = self.search_binary(higher) + 1

        result = []
        for i in range(lo, hi):
            mo = <MassObject>PyList_GetItem(self.structures, i)
            if lower <= mo.mass <= higher:
                result.append(mo)
        inst = self.__class__(
            [],
            self.mass_getter, sort=False)
        inst.structures = result
        return inst



import math
import numpy as np


sqrt2pi = np.sqrt(2 * np.pi)


def gauss(x, mu, sigma):
    return np.exp(-((x - mu) ** 2) / (2 * sigma ** 2)) / (sigma * sqrt2pi)


class MassAccuracyModel(object):

    @classmethod
    def fit(cls, matches):
        obs = [x.best_solution().precursor_mass_accuracy() for x in matches if len(x) > 0]
        mu = np.mean(obs)
        sigma = np.std(obs)
        return cls(mu, sigma)

    def __init__(self, mu, sigma, scale=1e6):
        self.mu = mu
        self.sigma = sigma
        self.scale = scale
        self.mu *= self.scale
        self.sigma *= self.scale

    def _sample(self):
        return self.score(self._interval(n_std=3))

    def _interval(self, n_std=3):
        w = self.sigma * n_std
        return np.arange(self.mu - w, self.mu + w, 1) / self.scale

    def score(self, error):
        return gauss(self.scale * error, self.mu, self.sigma)

    def __call__(self, error):
        return self.score(error)

    def __repr__(self):
        return "MassAccuracyModel(%e, %e)" % (self.mu / self.scale, self.sigma / self.scale)


class MassAccuracyMixin(object):
    accuracy_bias = MassAccuracyModel(0, 5e-6)

    def _precursor_mass_accuracy_score(self):
        offset, error = self.determine_precursor_offset(include_error=True)
        mass_accuracy = -10 * math.log10(1 - self.accuracy_bias(error))
        return mass_accuracy


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import gauss
except ImportError:
    pass



cimport cython

from libc.math cimport fabs

from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase
from glycresoft._c.database.mass_collection cimport NeutralMassDatabaseImpl


cdef bint isclose(double x, double y):
    return abs(x - y) < 1e-3


@cython.freelist(10000)
cdef class PeptideMassNode:
    cdef:
        public double peptide_mass
        public MassShiftBase mass_shift
        public double delta_mass
        public str scan_id
        public float glycan_score

    def __init__(self, peptide_mass, mass_shift, delta_mass, scan_id, glycan_score):
        self.peptide_mass = peptide_mass
        self.mass_shift = mass_shift
        self.delta_mass = delta_mass
        self.scan_id = scan_id
        self.glycan_score = glycan_score

    def __repr__(self):
        return f"{self.__class__.__name__}({self.peptide_mass}, {self.mass_shift}, {self.scan_id}, {self.glycan_score})"

    def __eq__(self, other):
        if other is None:
            return False
        if not isclose(self.peptide_mass, other.peptide_mass):
            return False
        if not isclose(self.delta_mass, other.delta_mass):
            return False
        if self.scan_id != other.scan_id:
            return False
        if self.mass_shift != other.mass_shift:
            return False

        return True

    def __ne__(self, other):
        return not self == other


@cython.cdivision(True)
cpdef list intersecting_peptide_masses(NeutralMassDatabaseImpl query_nodes,
                                       NeutralMassDatabaseImpl reference_nodes,
                                       float error_tolerance=5e-6):
    cdef:
        list accumulator
        size_t query_i, reference_i, query_n, reference_n
        size_t checkpoint
        PeptideMassNode query_node, reference_node

    accumulator = []

    query_i = 0
    reference_i = 0
    query_n = len(query_nodes)
    reference_n = len(reference_nodes)

    checkpoint = -1

    while query_i < query_n and reference_i < reference_n:
        query_node = <PeptideMassNode>query_nodes.get_item(query_i)
        reference_node = <PeptideMassNode>reference_nodes.get_item(reference_i)
        if fabs(query_node.peptide_mass - reference_node.peptide_mass) / reference_node.peptide_mass <= error_tolerance:
            accumulator.append((query_node, reference_node))
            if checkpoint == -1:
                checkpoint = reference_i
            query_i += 1
            reference_i += 1
        elif query_node.peptide_mass < reference_node.peptide_mass:
            if checkpoint != -1 and query_i < (query_n - 1) and fabs(query_node.peptide_mass - (
                    <PeptideMassNode>query_nodes.get_item(query_i + 1)).peptide_mass) / query_node.peptide_mass < error_tolerance:
                reference_i = checkpoint
                checkpoint = -1
            query_i += 1
        else:
            reference_i += 1
    return accumulator




from glycresoft.composition_distribution_model.constants import (
    DEFAULT_LAPLACIAN_REGULARIZATION,
    DEFAULT_RHO,
    RESET_THRESHOLD_VALUE,
    NORMALIZATION)

from glycresoft.composition_distribution_model.graph import (
    laplacian_matrix,
    adjacency_matrix,
    degree_matrix,
    weighted_laplacian_matrix,
    weighted_adjacency_matrix,
    weighted_degree_matrix,
    BlockLaplacian,
    network_indices,
    scale_network,
    assign_network,
    make_blocks)

from glycresoft.composition_distribution_model.laplacian_smoothing import (
    LaplacianSmoothingModel,
    ProportionMatrixNormalization,
    GroupBelongingnessMatrix,
    MatrixEditIndex,
    MatrixEditInstruction,
    BelongingnessMatrixPatcher)

from glycresoft.composition_distribution_model.grid_search import (
    NetworkReduction,
    NetworkTrimmingSearchSolution,
    GridSearchSolution,
    GridPointSolution,
    ThresholdSelectionGridSearch)


from glycresoft.composition_distribution_model.observation import (
    GlycanCompositionSolutionRecord,
    VariableObservationAggregation,
    AbundanceWeightedObservationAggregation,
    ObservationWeightState)


from glycresoft.composition_distribution_model.glycome_network_smoothing import (
    GlycomeModel,
    smooth_network)


from glycresoft.composition_distribution_model.utils import display_table


__all__ = [
    "DEFAULT_LAPLACIAN_REGULARIZATION",
    "DEFAULT_RHO",
    "RESET_THRESHOLD_VALUE",
    "NORMALIZATION",
    "laplacian_matrix",
    "adjacency_matrix",
    "degree_matrix",
    "weighted_laplacian_matrix",
    "weighted_adjacency_matrix",
    "weighted_degree_matrix",
    "BlockLaplacian",
    "network_indices",
    "scale_network",
    "assign_network",
    "make_blocks",
    "LaplacianSmoothingModel",
    "ProportionMatrixNormalization",
    "GroupBelongingnessMatrix",
    "MatrixEditIndex",
    "MatrixEditInstruction",
    "BelongingnessMatrixPatcher",
    "NetworkReduction",
    "NetworkTrimmingSearchSolution",
    "GridSearchSolution",
    "GridPointSolution",
    "ThresholdSelectionGridSearch",
    "GlycomeModel",
    "GlycanCompositionSolutionRecord",
    "VariableObservationAggregation",
    "AbundanceWeightedObservationAggregation",
    "NeighborhoodPrior",
    "smooth_network",
    "display_table",
    "ObservationWeightState"
]



import warnings
import io

from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Union


from glycopeptidepy.structure.parser import strip_modifications
from glycopeptidepy.utils.collectiontools import groupby

from glypy.structure.glycan_composition import HashableGlycanComposition

from glycresoft import serialize
from glycresoft.structure import PeptideProteinRelation, FragmentCachingGlycopeptide
from glycresoft.database.builder.glycopeptide.proteomics.fasta import DeflineSuffix
from glycresoft.database.builder.glycopeptide.proteomics.sequence_tree import SuffixTree


from .glycosite_model import MINIMUM, GlycosylationSiteModel


class ProteinStub(object):
    __slots__ = ('name', 'id', 'glycosylation_sites')

    name: str
    id: int
    glycosylation_sites: List[int]

    def __init__(self, name, id=None, glycosylation_sites=None):
        self.name = name
        self.id = id
        self.glycosylation_sites = glycosylation_sites

    def __getstate__(self):
        out = {
            'name': self.name,
            'id': self.id,
            'glycosylation_sites': self.glycosylation_sites
        }
        return out

    def __setstate__(self, state):
        self.name = state['name']
        self.id = state['id']
        self.glycosylation_sites = state['glycosylation_sites']

    def __repr__(self):
        return "{self.__class__.__name__}({self.name!r}, {self.id!r}, {self.glycosylation_sites!r})".format(self=self)

    @classmethod
    def from_protein(cls, protein):
        return cls(protein.name, protein.id)


def _make_name_suffix_lookup(proteins: Iterable[ProteinStub]):
    tree = SuffixTree()
    for prot in proteins:
        name = prot.name
        tree.add_ngram(DeflineSuffix(name, name))
    return tree


class GlycoproteinSiteSpecificGlycomeModel(object):
    __slots__ = ('protein', '_glycosylation_sites')

    protein: ProteinStub
    _glycosylation_sites: List[GlycosylationSiteModel]

    def __init__(self, protein, glycosylation_sites=None):
        self.protein = protein
        self._glycosylation_sites = []
        self.glycosylation_sites = glycosylation_sites

    def __getstate__(self):
        return {
            "protein": self.protein,
            "_glycosylation_sites": self._glycosylation_sites
        }

    def __setstate__(self, state):
        self.protein = state['protein']
        self._glycosylation_sites = state['_glycosylation_sites']

    def __eq__(self, other):
        if other is None:
            return False
        return (self.protein == other.protein) and (self.glycosylation_sites == other.glycosylation_sites)

    def __ne__(self, other):
        return not self == other

    @property
    def glycosylation_sites(self) -> List[GlycosylationSiteModel]:
        return self._glycosylation_sites

    @glycosylation_sites.setter
    def glycosylation_sites(self, glycosylation_sites):
        self._glycosylation_sites = sorted(
            glycosylation_sites or [], key=lambda x: x.position)

    def __getitem__(self, i) -> GlycosylationSiteModel:
        return self.glycosylation_sites[i]

    def __len__(self):
        return len(self.glycosylation_sites)

    @property
    def id(self):
        return self.protein.id

    @property
    def name(self):
        return self.protein.name

    def find_sites_in(self, start: int, end: int) -> List[GlycosylationSiteModel]:
        spans = []
        for site in self.glycosylation_sites:
            if start <= site.position <= end:
                spans.append(site)
            elif end < site.position:
                break
        return spans

    def _guess_sites_from_sequence(self, sequence: str):
        prot_seq = str(self.protein)
        query_seq = strip_modifications(sequence)
        try:
            start = prot_seq.index(query_seq)
            end = start + len(query_seq)
            return PeptideProteinRelation(start, end, self.protein.id, self.protein.hypothesis_id)
        except ValueError:
            return None

    def score(self, glycopeptide: FragmentCachingGlycopeptide,
              glycan_composition: Optional[HashableGlycanComposition]=None) -> float:
        if glycan_composition is None:
            glycan_composition = glycopeptide.glycan_composition
        pr = glycopeptide.protein_relation
        sites = self.find_sites_in(pr.start_position, pr.end_position)
        if len(sites) > 1:
            score_acc = 0.0
            warnings.warn(
                "Multiple glycosylation sites are not (yet) supported. Using average instead")
            for site in sites:
                try:
                    rec = site.glycan_map[glycan_composition]
                    score_acc += (rec.score)
                except KeyError:
                    score_acc += (MINIMUM)
            return score_acc / len(sites)
        try:
            site = sites[0]
            try:
                rec = site.glycan_map[glycan_composition]
            except KeyError:
                return MINIMUM
            return rec.score
        except IndexError:
            return MINIMUM

    @classmethod
    def bind_to_hypothesis(cls, session,
                           site_models: List[GlycosylationSiteModel],
                           hypothesis_id: int=1,
                           fuzzy: bool=True) -> Dict[int, 'GlycoproteinSiteSpecificGlycomeModel']:
        by_protein_name = defaultdict(list)
        for site in site_models:
            by_protein_name[site.protein_name].append(site)
        protein_models = {}
        proteins = session.query(serialize.Protein).filter(
            serialize.Protein.hypothesis_id == hypothesis_id).all()
        protein_name_map = {prot.name: prot for prot in proteins}
        tree = None
        for protein_name, sites in by_protein_name.items():
            try:
                protein = protein_name_map[protein_name]
            except KeyError:
                if fuzzy:
                    if tree is None:
                        tree = _make_name_suffix_lookup(proteins)
                    labels = list(tree.subsequences_of(protein_name))
                    if not labels:
                        continue
                    protein = protein_name_map[labels[0].original]
                else:
                    continue
            model = cls(protein, sites)
            protein_models[model.id] = model
        return protein_models

    def __repr__(self):
        template = "{self.__class__.__name__}({self.name}, {self.glycosylation_sites})"
        return template.format(self=self)

    @classmethod
    def load(cls, fh: io.TextIOBase,
             session=None,
             hypothesis_id: int = 1,
             fuzzy: bool = True) -> Union[List['GlycoproteinSiteSpecificGlycomeModel'],
                                          Dict[int, 'GlycoproteinSiteSpecificGlycomeModel']]:
        site_models = GlycosylationSiteModel.load(fh)
        if session is not None:
            return cls.bind_to_hypothesis(session, site_models, hypothesis_id, fuzzy)
        by_protein_name = groupby(site_models, lambda x: x.protein_name)
        result = []
        for name, models in by_protein_name.items():
            result.append(cls(ProteinStub(name), models))
        return result

    def stub_protein(self):
        self.protein = ProteinStub.from_protein(self.protein)



class ReversedProteinSiteReflectionGlycoproteinSiteSpecificGlycomeModel(GlycoproteinSiteSpecificGlycomeModel):
    __slots__ = ()

    @property
    def glycosylation_sites(self):
        return self._glycosylation_sites

    @glycosylation_sites.setter
    def glycosylation_sites(self, glycosylation_sites: List[GlycosylationSiteModel]):
        temp = []
        if isinstance(self.protein, ProteinStub):
            raise TypeError("Cannot create a reflected glycosite model with a ProteinStub")
        n = len(str(self.protein))
        for site in glycosylation_sites:
            site = site.copy()
            site.position = n - site.position - 1
            temp.append(site)
        self._glycosylation_sites = sorted(
            temp or [], key=lambda x: x.position)



import numpy as np

from glycresoft.task import TaskBase


class PriorBuilder(object):

    def __init__(self, network):
        self.network = network
        self.value = np.zeros(len(network))

    def bias(self, target, v):
        ix = self.network[target].index
        self.value[ix] += v
        return self

    def common(self, v):
        self.value += v
        return self

    def __getitem__(self, i):
        ix = self.network[i].index
        return self.value[ix]


class CompositionDistributionFit(object):

    def __init__(self, pi, network):
        self.pi = pi
        self.network = network

    def index_of(self, composition):
        return self.network[composition].index

    def __getitem__(self, composition):
        return self.pi[self.index_of(composition)]

    def __iter__(self):
        for i in range(len(self)):
            yield self.network[i], self.pi[i]

    def __len__(self):
        return len(self.network)

    def __repr__(self):
        return "CompositionDistributionFit()"


class CompositionDistributionUpdater(TaskBase):
    etol = 1e-6

    def __init__(self, network, solutions, prior=0.01, common_weight=1., precision_scale=1.):
        self.network = network
        self.solutions = solutions
        self.common_weight = common_weight
        self.precision_scale = precision_scale
        if isinstance(prior, (int, float)):
            prior = np.ones(len(network)) * prior
        self.prior = prior * precision_scale + common_weight

        # Dirichlet Mean
        self.pi0 = self.prior / self.prior.sum()

        self.sol_map = {
            sol.composition: sol.score for sol in solutions if sol.composition is not None}

        self.observations = np.array(
            [self.sol_map.get(node.composition.serialize(), 0) for node in network.nodes])
        self.iterations = 0

    def index_of(self, composition):
        return self.network[composition].index

    def convergence(self, pi_next, pi_last):
        d = np.abs(pi_last).sum()
        v = (np.abs(pi_next - pi_last).sum()) / d
        return v

    def update_pi(self, pi_array):
        pi2 = np.zeros_like(pi_array)
        total_score_pi = (self.observations * pi_array).sum()
        total_w = ((self.prior - 1).sum() + 1)
        for i in range(len(self.prior)):
            pi2[i] = ((self.prior[i] - 1) + (self.observations[i] *
                                             pi_array[i]) / total_score_pi) / total_w
            assert pi2[i] >= 0, (self.prior[i], pi_array[i])
        return pi2

    def optimize_pi(self, pi, maxiter=100, **kwargs):
        pi_last = pi
        pi_next = self.update_pi(pi_last)

        self.iterations = 0
        converging = float('inf')
        while converging > self.etol and self.iterations < maxiter:
            self.iterations += 1
            if self.iterations % 100 == 0:
                self.log("%f, %d" % (converging, self.iterations))
            pi_last = pi_next
            pi_next = self.update_pi(pi_last)
            converging = self.convergence(pi_next, pi_last)
        if converging < self.etol:
            self.log("Converged in %d iterations" % self.iterations)
        else:
            self.log("Failed to converge in %d iterations (%f)" %
                     (self.iterations, converging))

        return pi_next

    def fit(self, **kwargs):
        return CompositionDistributionFit(
            self.optimize_pi(self.pi0, **kwargs), self.network)



import os
from typing import Optional
from glycresoft.serialize.hypothesis.peptide import (Peptide, Glycopeptide)

from .common import (
    GlycopeptideHypothesisSerializerBase, PeptideGlycosylator,
    MultipleProcessPeptideGlycosylator)
from .proteomics import mzid_proteome


class MzIdentMLGlycopeptideHypothesisSerializer(GlycopeptideHypothesisSerializerBase):
    _display_name = "MzIdentML Glycopeptide Hypothesis Serializer"

    def __init__(self, mzid_path, connection, glycan_hypothesis_id, hypothesis_name=None,
                 target_proteins=None, max_glycosylation_events=1, reference_fasta=None,
                 full_cross_product=True, peptide_length_range=(5, 60), use_uniprot=True,
                 uniprot_source_file: Optional[str]=None):
        if target_proteins is None:
            target_proteins = []
        GlycopeptideHypothesisSerializerBase.__init__(
            self, connection, hypothesis_name, glycan_hypothesis_id,
            full_cross_product, use_uniprot=use_uniprot, uniprot_source_file=uniprot_source_file)
        self.mzid_path = mzid_path
        self.reference_fasta = reference_fasta
        self._make_proteome(mzid_path, target_proteins, reference_fasta)
        self.target_proteins = target_proteins
        self.max_glycosylation_events = max_glycosylation_events
        self.peptide_length_range = peptide_length_range or (5, 60)
        assert len(self.peptide_length_range) == 2

        self.set_parameters({
            "mzid_file": os.path.abspath(mzid_path),
            "reference_fasta": os.path.abspath(
                reference_fasta) if reference_fasta is not None else None,
            "target_proteins": target_proteins,
            "max_glycosylation_events": max_glycosylation_events,
            "full_cross_product": self.full_cross_product,
            "peptide_length_range": self.peptide_length_range
        })

    def _make_proteome(self, mzid_path, target_proteins, reference_fasta):
        self.proteome = mzid_proteome.Proteome(
            mzid_path, self._original_connection, self.hypothesis_id,
            target_proteins=target_proteins,
            reference_fasta=reference_fasta,
            use_uniprot=self.use_uniprot,
            uniprot_source_file=self.uniprot_source_file
        )

    def retrieve_target_protein_ids(self):
        return self.proteome.retrieve_target_protein_ids()

    def peptide_ids(self):
        out = []
        for protein_id in self.retrieve_target_protein_ids():
            out.extend(i[0] for i in self.query(Peptide.id).filter(
                Peptide.protein_id == protein_id))
        return out

    def glycosylate_peptides(self):
        glycosylator = PeptideGlycosylator(self.session, self.hypothesis_id)
        acc = []
        i = 0
        for peptide_id in self.peptide_ids():
            peptide = self.query(Peptide).get(peptide_id)
            for glycopeptide in glycosylator.handle_peptide(peptide):
                acc.append(glycopeptide)
                i += 1
                if len(acc) > 100000:
                    self.session.bulk_insert_mappings(Glycopeptide, acc, render_nulls=True)
                    self.session.commit()
                    acc = []
        self.session.bulk_insert_mappings(Glycopeptide, acc, render_nulls=True)
        self.session.commit()

    def run(self):
        self.log("Loading Proteome")
        self.proteome.load()
        self.set_parameters({
            "enzymes": self.proteome.enzymes,
            "constant_modifications": self.proteome.constant_modifications,
            "include_baseline_peptides": self.proteome.include_baseline_peptides
        })
        self.log("Combinating Glycans")
        self.combinate_glycans(self.max_glycosylation_events)
        if self.full_cross_product:
            self.log("Building Glycopeptides")
            self.glycosylate_peptides()
        self._sql_analyze_database()
        if self.full_cross_product:
            self._count_produced_glycopeptides()
        self.log("Done")


class MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer(MzIdentMLGlycopeptideHypothesisSerializer):
    _display_name = "Multiple Process MzIdentML Glycopeptide Hypothesis Serializer"

    def __init__(self, mzid_path, connection, glycan_hypothesis_id, hypothesis_name=None,
                 target_proteins=None, max_glycosylation_events=1, reference_fasta=None,
                 full_cross_product=True, peptide_length_range=(5, 60),
                 uniprot_source_file: Optional[str]=None,
                 n_processes=4):
        super(MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer, self).__init__(
            mzid_path, connection, glycan_hypothesis_id, hypothesis_name, target_proteins,
            max_glycosylation_events, reference_fasta, full_cross_product,
            peptide_length_range, uniprot_source_file=uniprot_source_file)
        self.n_processes = n_processes
        self.proteome.n_processes = n_processes

    def glycosylate_peptides(self):
        dispatcher = MultipleProcessPeptideGlycosylator(
            self._original_connection, self.hypothesis_id,
            glycan_combination_count=self.total_glycan_combination_count,
            n_processes=self.n_processes)
        dispatcher.process(self.peptide_ids())


class ReversingMultipleProcessMzIdentMLGlycopeptideHypothesisSerializer(MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer):
    def _make_proteome(self, mzid_path, target_proteins, reference_fasta):
        self.proteome = mzid_proteome.ReverseProteome(
            mzid_path, self._original_connection, self.hypothesis_id,
            target_proteins=target_proteins,
            reference_fasta=reference_fasta,
            use_uniprot=self.use_uniprot,
            uniprot_source_file=self.uniprot_source_file
        )


class MzIdentMLPeptideHypothesisSerializer(GlycopeptideHypothesisSerializerBase):
    def __init__(self, mzid_path, connection, hypothesis_name=None,
                 target_proteins=None, reference_fasta=None,
                 include_baseline_peptides=False,
                 uniprot_source_file: Optional[str] = None):
        if target_proteins is None:
            target_proteins = []
        GlycopeptideHypothesisSerializerBase.__init__(
            self, connection, hypothesis_name, 0, uniprot_source_file=uniprot_source_file)
        self.mzid_path = mzid_path
        self.reference_fasta = reference_fasta
        self.proteome = mzid_proteome.Proteome(
            mzid_path, self._original_connection, self.hypothesis_id,
            target_proteins=target_proteins, reference_fasta=reference_fasta,
            include_baseline_peptides=include_baseline_peptides)
        self.target_proteins = target_proteins
        self.include_baseline_peptides = include_baseline_peptides

        self.set_parameters({
            "mzid_file": os.path.abspath(mzid_path),
            "reference_fasta": os.path.abspath(
                reference_fasta) if reference_fasta is not None else None,
            "target_proteins": target_proteins,
        })

    def retrieve_target_protein_ids(self):
        # if len(self.target_proteins) == 0:
        #     return [
        #         i[0] for i in
        #         self.query(Protein.id).filter(
        #             Protein.hypothesis_id == self.hypothesis_id).all()
        #     ]
        # else:
        #     result = []
        #     for target in self.target_proteins:
        #         if isinstance(target, basestring):
        #             if target in self.proteome.accession_map:
        #                 target = self.proteome.accession_map[target]
        #             match = self.query(Protein.id).filter(
        #                 Protein.name == target,
        #                 Protein.hypothesis_id == self.hypothesis_id).first()
        #             if match:
        #                 result.append(match[0])
        #             else:
        #                 self.log("Could not locate protein '%s'" % target)
        #         elif isinstance(target, int):
        #             result.append(target)
        #     return result
        return self.proteome.retrieve_target_protein_ids()

    def peptide_ids(self):
        out = []
        for protein_id in self.retrieve_target_protein_ids():
            out.extend(i[0] for i in self.query(Peptide.id).filter(
                Peptide.protein_id == protein_id))
        return out

    def run(self):
        self.log("Loading Proteome")
        self.proteome.load()
        self.set_parameters({
            "enzymes": self.proteome.enzymes,
            "constant_modifications": self.proteome.constant_modifications,
            "include_baseline_peptides": self.proteome.include_baseline_peptides
        })
        self.log("Done")



import csv

import io
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, Mapping
from glycresoft.structure.structure_loader import GlycanCompositionDeltaCache

import numpy as np
from scipy.ndimage import gaussian_filter1d, median_filter

try:
    from matplotlib import pyplot as plt
except ImportError:
    pass

from glycopeptidepy import PeptideSequence

from glypy.structure.glycan_composition import HashableGlycanComposition, FrozenGlycanComposition

from glycresoft.structure import FragmentCachingGlycopeptide
from ms_deisotope.peak_dependency_network.intervals import SpanningMixin, IntervalTreeNode

from glycresoft.chromatogram_tree.utils import ArithmeticMapping
from glycresoft.chromatogram_tree.mass_shift import MassShiftBase, Unmodified
from glycresoft.chromatogram_tree.chromatogram import ChromatogramInterface

from glycresoft.tandem.spectrum_match import SpectrumMatchBase


def _try_parse(value):
    if isinstance(value, (int, float)):
        return value
    try:
        return int(value)
    except (ValueError, TypeError):
        try:
            return float(value)
        except (ValueError, TypeError):
            return value


def _get_apex_time(chromatogram):
    try:
        if hasattr(chromatogram, "as_arrays"):
            x, y = chromatogram.as_arrays()
        else:
            x, y = chromatogram.get_chromatogram().as_arrays()
        y = gaussian_filter1d(y, 1)
        return x[np.argmax(y)]
    except (AttributeError, TypeError):
        return chromatogram.apex_time


class ChromatogramProxy(object):

    weighted_neutral_mass: float
    apex_time: float
    total_signal: float
    glycan_composition: HashableGlycanComposition
    source: Any
    mass_shifts: List[MassShiftBase]
    weight: float
    kwargs: Dict[str, Any]

    def __init__(self, weighted_neutral_mass, apex_time, total_signal, glycan_composition, source=None,
                 mass_shifts=None, weight=1.0, **kwargs):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        self.weighted_neutral_mass = weighted_neutral_mass
        self.apex_time = apex_time
        self.total_signal = total_signal
        self.glycan_composition = glycan_composition
        self.source = source
        self.mass_shifts = mass_shifts
        self.weight = weight
        self.kwargs = kwargs

    @property
    def revised_from(self):
        return self.kwargs.get("revised_from")

    @revised_from.setter
    def revised_from(self, value):
        self.kwargs['revised_from'] = value

    @property
    def annotations(self):
        return self.kwargs

    @property
    def tag(self) -> Optional[int]:
        return self.kwargs.get('tag')

    @tag.setter
    def tag(self, value: int):
        self.kwargs['tag'] = value

    def __repr__(self):
        return "%s(%f, %f, %f, %s, %s)" % (
            self.__class__.__name__,
            self.weighted_neutral_mass, self.apex_time, self.total_signal,
            self.glycan_composition, self.kwargs)

    def pack(self):
        self.source = None

    @classmethod
    def from_obj(cls, obj, **kwargs):
        if isinstance(obj, ChromatogramInterface):
            return cls.from_chromatogram(obj)
        elif isinstance(obj, cls):
            return obj.__class__.from_chromatogram(obj, **kwargs)
        elif isinstance(obj, SpectrumMatchBase):
            return cls.from_spectrum_match(obj, **kwargs)
        else:
            return cls.from_chromatogram(obj)

    @classmethod
    def from_chromatogram(cls, obj, **kwargs):
        try:
            apex_time = _get_apex_time(obj)
        except (AttributeError, ValueError, TypeError):
            apex_time = obj.apex_time
        mass_shifts = getattr(obj, 'mass_shifts')
        kwargs.setdefault('mass_shifts', mass_shifts)
        inst = cls(
            obj.weighted_neutral_mass, apex_time, obj.total_signal,
            HashableGlycanComposition(obj.glycan_composition), obj, **kwargs)
        return inst

    @classmethod
    def from_spectrum_match(cls, spectrum_match, source=None, **kwargs):
        if source is None:
            source = spectrum_match
        return cls(
            spectrum_match.scan.precursor_information.neutral_mass, spectrum_match.scan.scan_time, 0.0,
            spectrum_match.target.glycan_composition, source, [spectrum_match.mass_shift], **kwargs)

    def get_chromatogram(self):
        return self.source.get_chromatogram()

    def _to_csv(self):
        d = self._prepare_state()
        d['mass_shifts'] = ';'.join(m.name for m in d['mass_shifts'])
        return d

    def _prepare_state(self):
        d = {
            "weighted_neutral_mass": self.weighted_neutral_mass,
            "apex_time": self.apex_time,
            "total_signal": self.total_signal,
            "glycan_composition": self.glycan_composition,
            "weight": float(self.weight),
            "mass_shifts": [m for m in self.mass_shifts],
        }
        d.update(self.kwargs)
        return d

    @classmethod
    def _from_state(cls, state):
        new = cls.__new__(cls)
        new.__setstate__(state)
        return new

    def copy(self):
        dup = self._from_state(self._prepare_state())
        dup.source = self.source
        return dup

    def __eq__(self, other: 'ChromatogramProxy'):
        if other is None:
            return False
        if self.glycan_composition != other.glycan_composition:
            return False
        if not np.isclose(self.apex_time, other.apex_time):
            return False
        if not np.isclose(self.total_signal, other.total_signal):
            return False
        if self.mass_shifts != other.mass_shifts:
            return False
        if not np.isclose(self.weighted_neutral_mass, other.weighted_neutral_mass):
            return False
        return True

    def __ne__(self, other: 'ChromatogramProxy'):
        return not self == other

    def __hash__(self):
        return hash(self.glycan_composition)

    def __getstate__(self):
        state = self._prepare_state()
        state['glycan_composition'] = str(state['glycan_composition'])
        state['weight'] = self.weight
        return state

    def __setstate__(self, state):
        state = dict(state)
        self.glycan_composition = state.pop("glycan_composition", None)
        if self.glycan_composition is not None:
            if isinstance(self.glycan_composition, HashableGlycanComposition):
                pass
            else:
                self.glycan_composition = HashableGlycanComposition.parse(str(self.glycan_composition))
        self.apex_time = state.pop("apex_time", None)
        self.total_signal = state.pop("total_signal", None)
        self.weighted_neutral_mass = state.pop("weighted_neutral_mass", None)
        self.weight = state.pop('weight', 1.0)
        self.source = None
        self.mass_shifts = state.pop("mass_shifts", [Unmodified])
        self.kwargs = state

    def __getattr__(self, attr):
        try:
            return self.annotations[attr]
        except KeyError:
            raise AttributeError(attr)

    @classmethod
    def _csv_keys(cls, keys):
        return ['glycan_composition', 'apex_time', 'total_signal', 'weighted_neutral_mass', ] + \
            sorted(set(keys) - {'glycan_composition', 'apex_time',
                                'total_signal', 'weighted_neutral_mass', })

    @classmethod
    def to_csv(cls, instances: Iterable, fh: io.TextIOBase):
        cases = [c._to_csv() for c in instances]
        keys = cls._csv_keys(cases[0].keys())
        writer = csv.DictWriter(fh, fieldnames=keys)
        writer.writeheader()
        writer.writerows(cases)

    @classmethod
    def _from_csv(cls, row):
        mass = float(row.pop("weighted_neutral_mass"))
        apex_time = float(row.pop("apex_time"))
        total_signal = float(row.pop("total_signal"))
        gc = HashableGlycanComposition.parse(row.pop("glycan_composition"))
        kwargs = {k: _try_parse(v) for k, v in row.items()}
        return cls(mass, apex_time, total_signal, gc, **kwargs)

    @classmethod
    def from_csv(cls, fh: io.TextIOBase):
        cases = []
        reader = csv.DictReader(fh)

        for row in reader:
            cases.append(cls._from_csv(row))
        return cases

    def update_glycan_composition(self, glycan_composition: HashableGlycanComposition) -> 'ChromatogramProxy':
        self.glycan_composition = glycan_composition

    def shift_glycan_composition(self, delta: HashableGlycanComposition):
        inst = self.copy()
        if isinstance(self.glycan_composition, HashableGlycanComposition):
            new_gc = self.glycan_composition + delta
        else:
            new_gc = HashableGlycanComposition(self.glycan_composition) + delta
        inst.update_glycan_composition(new_gc)
        return inst


class GlycopeptideChromatogramProxy(ChromatogramProxy):
    _structure = None

    @classmethod
    def _csv_keys(cls, keys):
        return ['structure', 'glycan_composition', 'apex_time',
                'total_signal', 'weighted_neutral_mass'] + \
            sorted(set(keys) - {'structure', 'glycan_composition',
                                'apex_time', 'total_signal', 'weighted_neutral_mass'})

    @property
    def peptide_key(self) -> str:
        if "peptide_key" in self.kwargs:
            return self.kwargs["peptide_key"]
        structure = self.structure
        if isinstance(structure, str):
            peptide = str(PeptideSequence(structure).deglycosylate())
        else:
            peptide = str(structure.clone().deglycosylate())
        self.kwargs["peptide_key"] = peptide
        return peptide

    @property
    def structure(self) -> FragmentCachingGlycopeptide:
        if self._structure is None:
            self._structure = FragmentCachingGlycopeptide(
                str(self.kwargs["structure"]))
        return self._structure

    @structure.setter
    def structure(self, value: FragmentCachingGlycopeptide):
        self._structure = value
        self.kwargs['structure'] = str(value)
        if value and self.glycan_composition != value.glycan_composition:
            self.glycan_composition = HashableGlycanComposition(
                value.glycan_composition)

    @classmethod
    def from_chromatogram(cls, obj, **kwargs):
        gp = FragmentCachingGlycopeptide(str(obj.structure))
        result = super(GlycopeptideChromatogramProxy, cls).from_chromatogram(
            obj, structure=gp, **kwargs)
        _key = result.peptide_key
        return result

    @classmethod
    def from_spectrum_match(cls, spectrum_match, source=None, **kwargs):
        if source is None:
            source = spectrum_match
        gp = FragmentCachingGlycopeptide(str(spectrum_match.target))
        return cls(
            spectrum_match.scan.precursor_information.neutral_mass, spectrum_match.scan.scan_time, 0.0,
            spectrum_match.target.glycan_composition, source, [spectrum_match.mass_shift], structure=gp, **kwargs)

    def shift_glycan_composition(self, delta):
        inst = super(GlycopeptideChromatogramProxy,
                     self).shift_glycan_composition(delta)
        return inst

    def update_glycan_composition(self, glycan_composition: HashableGlycanComposition) -> 'GlycopeptideChromatogramProxy':
        original_structure = str(self.structure)
        super().update_glycan_composition(glycan_composition)
        structure = self.structure.clone()
        structure.glycan = self.glycan_composition
        self.structure = structure
        self.kwargs['original_structure'] = original_structure

    def __eq__(self, other: 'GlycopeptideChromatogramProxy'):
        result = super().__eq__(other)
        if not result:
            return result
        return self.structure == other.structure

    def __hash__(self):
        return hash(self.structure)


class CommonGlycopeptide(object):
    def __init__(self, structure, mapping):
        self.mapping = mapping
        self.structure = structure

    def __getitem__(self, key):
        return self.mapping[key]

    def __setitem__(self, key, value):
        self.mapping[key] = value

    def __delitem__(self, key):
        del self.mapping[key]

    def keys(self):
        return self.mapping.keys()

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    @property
    def glycan_composition(self):
        return self.structure.glycan_composition

    @property
    def apex_times(self):
        return ArithmeticMapping({k: v.apex_time for k, v in self.mapping.items()})

    @property
    def total_signals(self):
        return ArithmeticMapping({k: v.total_signal for k, v in self.mapping.items()})

    def __repr__(self):
        template = "{self.__class__.__name__}({self.structure!s}, {count} observations)"
        count = len(self.mapping)
        return template.format(self=self, count=count)


class GlycoformAggregator(Mapping):
    _interval_tree: IntervalTreeNode
    by_peptide: Dict[str, 'GlycoformGroup']
    factors: Dict[str, float]

    def __init__(self, glycoforms=None, *args, **kwargs):
        self.by_peptide = dict()
        self.factors = None
        self._interval_tree = None
        if glycoforms is not None:
            self.gather(glycoforms)

    def _get_peptide_key(self, chromatogram):
        if isinstance(chromatogram, GlycopeptideChromatogramProxy):
            return chromatogram.peptide_key
        return str(PeptideSequence(str(chromatogram.structure)).deglycosylate())

    def gather(self, iterable: Iterable[GlycopeptideChromatogramProxy]):
        for rec in iterable:
            key = self._get_peptide_key(rec)
            try:
                group = self.by_peptide[key]
            except KeyError:
                group = self.by_peptide[key] = GlycoformGroup([], key)
            group.append(rec)
        self._reindex()
        return self

    def _reindex(self):
        self.factors = self._infer_factors(self.glycoforms())
        self._interval_tree = IntervalTreeNode.build(self.values())
        return self

    def overlaps(self, start: float, end: float) -> List[GlycopeptideChromatogramProxy]:
        groups = self._interval_tree.overlaps(start, end)
        out = []
        for group in groups:
            out.extend(group.overlaps(start, end))
        return out

    @property
    def start_time(self):
        return min(v.start_time for v in self.values())

    @property
    def end_time(self):
        return max(v.end_time for v in self.values())

    def __getitem__(self, key) -> 'GlycoformGroup':
        if not isinstance(key, str):
            key = self._get_peptide_key(key)
        return self.by_peptide[key]

    def __setitem__(self, key, value: 'GlycoformGroup'):
        if not isinstance(key, str):
            key = self._get_peptide_key(key)
        self.by_peptide[key] = value

    def __contains__(self, key) -> bool:
        if not isinstance(key, str):
            key = self._get_peptide_key(key)
        return key in self.by_peptide

    def keys(self):
        return (self.by_peptide).keys()

    def __iter__(self):
        return self.glycoforms()

    def values(self):
        return self.by_peptide.values()

    def items(self):
        return self.by_peptide.items()

    def __len__(self):
        return len(self.by_peptide)

    def has_relative_pairs(self) -> bool:
        return any(len(v) > 1 for k, v in self.by_peptide.items())

    def _deltas_for(self, monosaccharide: str, include_pairs: bool = False):
        deltas = []
        pairs = []
        for _backbone, cases in self.by_peptide.items():
            for target in cases:
                gc = FrozenGlycanComposition.parse(
                    str(target.glycan_composition))
                gc[monosaccharide] += 1
                for case in cases:
                    if case.glycan_composition == gc:
                        deltas.append(case.apex_time - target.apex_time)
                        if include_pairs:
                            pairs.append((case, target, deltas[-1]))
        deltas = np.array(deltas)
        if include_pairs:
            return deltas, pairs
        return deltas

    def _infer_factors(self, chromatograms: Iterable[GlycopeptideChromatogramProxy]) -> List[str]:
        keys = set()
        for record in chromatograms:
            keys.update(record.glycan_composition)
        keys = sorted(map(str, keys))
        return keys

    def deltas(self, factors: Iterable[str] = None, include_pairs: bool = False):
        if factors is None:
            factors = self.factors
        return {f: self._deltas_for(f, include_pairs=include_pairs) for f in factors}

    def glycoforms(self) -> Iterator[GlycopeptideChromatogramProxy]:
        for group in self.values():
            for member in group:
                yield member

    def tag(self):
        i = 0
        for member in self:
            member.tag = i
            i += 1
        return self

    def get_tag(self, tag: int) -> GlycopeptideChromatogramProxy:
        for x in self:
            if x.tag == tag:
                return x
        raise KeyError(tag)

    def has_tag(self, tag: int) -> bool:
        for x in self:
            if x.tag == tag:
                return True
        return False


class GlycoformGroup(Sequence[GlycopeptideChromatogramProxy], SpanningMixin):
    members: List[GlycopeptideChromatogramProxy]
    group_id: str
    start: float
    end: float
    base_time: Optional[float]

    def __init__(self, members, group_id):
        self.members = list(members)
        self.group_id = group_id
        self.start_time = 0
        self.end_time = 0
        self.base_time = None

    @property
    def start_time(self):
        return self.start

    @start_time.setter
    def start_time(self, value):
        self.start = value

    @property
    def end_time(self):
        return self.end

    @end_time.setter
    def end_time(self, value):
        self.end = value

    def overlaps(self, start: float, end: float):
        return [m for m in self if start <= m.apex_time <= end]

    def _update(self):
        start_time = float('inf')
        end_time = 0.0
        for member in self.members:
            if member.apex_time < start_time:
                start_time = member.apex_time
            if member.apex_time > end_time:
                end_time = member.apex_time
        self.start_time = start_time
        self.end_time = end_time

    def __len__(self):
        return len(self.members)

    def __getitem__(self, i):
        return self.members[i]

    def append(self, glycoform: GlycopeptideChromatogramProxy):
        self.members.append(glycoform)
        self._update()

    def extend(self, glycoforms: Iterable[GlycopeptideChromatogramProxy]):
        self.members.extend(glycoforms)
        self._update()

    def __repr__(self):
        return "{self.__class__.__name__}({self.members!r}, {self.group_id!r})".format(self=self)


class DeltaOverTimeFilter(object):
    def __init__(self, key, delta, time):
        self.key = key
        self.delta = np.array(delta)
        self.time = np.array(time)
        self.delta = self.gap_fill()
        self.smoothed = self.smooth()

    def smooth(self):
        return median_filter(self.delta, size=5, mode='nearest')

    def __array__(self):
        return self.smoothed

    def __getitem__(self, i):
        return self.smoothed[i]

    def __len__(self):
        return len(self.time)

    def gap_fill(self, width=2):
        x = self.delta.copy()
        for i in range(len(x)):
            if np.isnan(x[i]):
                x[i] = np.nanmean(np.concatenate(
                    (x[max(i - width, 0):i], x[i + 1:i + width])))
        return x

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        ax.plot(self.time, self.delta, label='Raw')
        ax.scatter(self.time, self.delta)
        ax.plot(self.time, self.smoothed,
                label="Smoothed", alpha=0.95, ls='--')
        return ax

    def search(self, time):
        lo = 0
        n = hi = len(self.time)
        while hi != lo:
            mid = (hi + lo) // 2
            x = self.time[mid]
            err = (x - time)
            if err == 0:
                return mid
            elif (hi - lo) == 1:
                err = abs(err)
                if mid != 0:
                    prev_err = self.time[mid - 1] - time
                    if abs(prev_err) < err:
                        return mid - 1
                if mid < n - 1:
                    next_err = self.time[mid + 1] - time
                    if abs(next_err) < err:
                        return mid + 1
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid



from ms_deisotope import CompositionListPeakDependenceGraphDeconvoluter, DistinctPatternFitter

from glycresoft.database.composition_aggregation import Formula, CommonComposition, Composition
from glycresoft.tandem.spectrum_match import SpectrumMatcherBase


from .score_components import IntensityRankScorer, FragmentScorer


def get_fragments(assigned_peaks):
    results = set()
    for peak in assigned_peaks:
        results.update(peak.solution.base_data)
    return results


class BinomialCoverageSpectrumMatcher(SpectrumMatcherBase):
    """Encapsulates the process of matching experimental peaks to theoretical peaks.

    Given a :class:`ms_peak_picker.PeakIndex` from a :class:`MSMSScan` and a :class:`StructureRecord`
    instance, this type will determine the set of all compositions needed to search, handling ambiguous
    compositions and managing the searching process.

    Attributes
    ----------
    assigned_peaks : list
        The list of deconvolved peaks that were assigned compositions
    composition_list : list
        The list of compositions to attempt to assign
    matched_fragments : list
        The list of theoretical fragments for which compositions were assigned
    peak_set : ms_peak_picker.PeakIndex
        The list of experimental peaks
    structure_record : StructureRecord
        The structure to search for fragments from
    """
    def __init__(self, scan, target):
        super(BinomialCoverageSpectrumMatcher, self).__init__(scan, target)
        self.peak_set = self.scan.peak_set
        self.composition_list = None
        self.assigned_peaks = None
        self.matched_fragments = None

        self.make_composition_list()

    @property
    def structure(self):
        return self.target

    def make_composition_list(self):
        """Aggregate all theoretical compositions by formula so that each composition
        is present exactly once, containing all possible sources of that composition in
        a single :class:`CommonComposition` instance.

        Sets the :attr:`composition_list` attribute. Called during initialization.
        """
        self.composition_list = CommonComposition.aggregate(
            Formula(f.composition, f) for f in self.structure.fragment_map.values())

        # Is there anything added by including the precursor? Maybe in the graph solving step
        # f = Formula(self.structure_record.structure.total_composition(), self.structure_record)
        # self.composition_list.append(CommonComposition(f, [f.data]))

    def match(self, mass_tolerance=5e-6, charge_carrier=Composition("Na").mass, fitter_parameters=None):
        """Perform the actual matching of peaks, fitting isotopic patterns and retrieving the matched
        structural features.

        Parameters
        ----------
        mass_tolerance : float, optional
            The parts-per-million mass error tolerance allowed. Defaults to 5ppm, entered as 5e-6
        charge_range : tuple, optional
            The a tuple defining the minimum and maximum of the range of charge states to consider
            for each composition. Defaults to positive (1, 3)
        charge_carrier : float, optional
            The mass of the charge carrying element. By default for the experimental set up described
            here, this is the mass of a sodium (Na)

        Returns
        -------
        SpectrumSolution
        """

        if fitter_parameters is None:
            fitter_parameters = {}

        charge_range = (1, self.scan.precursor_information.charge)
        dec = CompositionListPeakDependenceGraphDeconvoluter(
            self.peak_set, self.composition_list,
            scorer=DistinctPatternFitter(**fitter_parameters))
        self.assigned_peaks = dec.deconvolute(
            charge_range=charge_range, charge_carrier=charge_carrier,
            error_tolerance=mass_tolerance)
        self.matched_fragments = list(get_fragments(self.assigned_peaks))

        self._structure_scorer = FragmentScorer(self.structure, self.matched_fragments)
        self._spectrum_scorer = IntensityRankScorer(self.peak_set, self.assigned_peaks)
        self.score = self._spectrum_scorer.final_score + self._structure_scorer.final_score



# -*- coding: utf-8 -*-
import math
import logging

from typing import Callable, List, NamedTuple, Optional, Sequence, Tuple, Union
from collections import defaultdict, namedtuple

import numpy as np
try:
    from matplotlib import pyplot as plt
except (ImportError, RuntimeError):
    plt = None

from glycresoft.task import LoggingMixin
from glycresoft.tandem.spectrum_match import SpectrumSolutionSet, SpectrumMatch

ScoreCell = namedtuple('ScoreCell', ['score', 'value'])

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class NearestValueLookUp(object):
    """
    A mapping-like object which simplifies
    finding the value of a pair whose key is nearest
    to a given query.

    .. note::
        Queries exceeding the maximum key will return
        the maximum key's value.
    """

    def __init__(self, items):
        if isinstance(items, dict):
            items = items.items()
        self.items = sorted(
            [ScoreCell(*x) for x in items if not np.isnan(x[0])], key=lambda x: x[0])

    def max_key(self):
        try:
            return self.items[-1][0]
        except IndexError:
            return 0

    def _find_closest_item(self, value, key_index=0):
        array = self.items
        lo = 0
        hi = len(array)
        n = hi

        error_tolerance = 1e-3

        if np.isnan(value):
            return lo

        if lo == hi:
            return lo

        while hi - lo:
            i = (hi + lo) // 2
            x = array[i][key_index]
            err = x - value
            if abs(err) < error_tolerance:
                mid = i
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = array[i][key_index]
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = array[i][key_index]
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif (hi - lo) == 1:
                mid = i
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = array[i][key_index]
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = array[i][key_index]
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif x < value:
                lo = i
            elif x > value:
                hi = i

    def get_pair(self, key, key_index=0):
        k = self._find_closest_item(key, key_index) + 1
        return self.items[k]

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return "{s.__class__.__name__}({size})".format(
            s=self, size=len(self))

    def __getitem__(self, key):
        return self._get_one(key)

    def _get_sequence(self, key):
        value = [self._get_one(k) for k in key]
        if isinstance(key, np.ndarray):
            value = np.array(value, dtype=float)
        return value

    def _get_one(self, key):
        ix = self._find_closest_item(key)
        if ix >= len(self):
            ix = len(self) - 1
        if ix < 0:
            ix = 0
        pair = self.items[ix]
        return pair[1]


try:
    _NearestValueLookUp = NearestValueLookUp
    from glycresoft._c.tandem.target_decoy import NearestValueLookUp as NearestValueLookUp
except ImportError:
    pass


class ScoreThresholdCounter(object):
    def __init__(self, series, thresholds):
        self.series = self._prepare_series(series)
        self.thresholds = sorted(set(np.round((thresholds), 10)))
        self.counter = defaultdict(int)
        self.counts_above_threshold = None
        self.n_thresholds = len(self.thresholds)
        self.threshold_index = 0
        self.current_threshold = self.thresholds[self.threshold_index]
        self.current_count = 0

        self._i = 0
        self._is_done = False

        self.find_counts()
        self.counts_above_threshold = self.compute_complement()
        self.counter = NearestValueLookUp(self.counter)

    def _prepare_series(self, series):
        return sorted(series, key=lambda x: x.score)

    def advance_threshold(self):
        self.threshold_index += 1
        if self.threshold_index < self.n_thresholds:
            self.current_threshold = self.thresholds[self.threshold_index]
            self.counter[self.current_threshold] = self.current_count
            return True
        else:
            self._is_done = True
            return False

    def test(self, item):
        if item.score < self.current_threshold:
            self.current_count += 1
            self._i += 1
        else:
            # Rather than using recursion, just invert the condition
            # being tested and loop here.
            while self.advance_threshold():
                if item.score > self.current_threshold:
                    continue
                else:
                    self.current_count += 1
                    self._i += 1
                    break

    def find_counts(self):
        for item in self.series:
            self.test(item)

    def compute_complement(self):
        complement = defaultdict(int)
        n = len(self.series)

        for k, v in self.counter.items():
            complement[k] = n - v
        return NearestValueLookUp(complement)


class ArrayScoreThresholdCounter(ScoreThresholdCounter):

    def _prepare_series(self, series):
        return np.sort(np.array(series))

    def test(self, item):
        if item < self.current_threshold:
            self.current_count += 1
            self._i += 1
        else:
            # Rather than using recursion, just invert the condition
            # being tested and loop here.
            while self.advance_threshold():
                if item > self.current_threshold:
                    continue
                else:
                    self.current_count += 1
                    self._i += 1
                    break


class TargetDecoySet(NamedTuple):
    target_matches: List[SpectrumSolutionSet]
    decoy_matches: List[SpectrumSolutionSet]

    def target_count(self):
        return len(self.target_matches)

    def decoy_count(self):
        return len(self.decoy_matches)


# implementation derived from pyteomics
_precalc_fact = np.log([math.factorial(n) for n in range(20)])


def log_factorial(x):
    x = np.array(x)
    m = (x >= _precalc_fact.size)
    out = np.empty(x.shape)
    out[~m] = _precalc_fact[x[~m].astype(int)]
    x = x[m]
    out[m] = x * np.log(x) - x + 0.5 * np.log(2 * np.pi * x)
    return out


def _log_pi_r(d, k, p=0.5):
    return k * math.log(p) + log_factorial(k + d) - log_factorial(k) - log_factorial(d)


def _log_pi(d, k, p=0.5):
    return _log_pi_r(d, k, p) + (d + 1) * math.log(1 - p)


def _expectation(d, t, p=0.5):
    """The conditional tail probability for the negative binomial
    random variable for the number of incorrect target matches

    Parameters
    ----------
    d : int
        The number of decoys retained
    t : int
        The number of targets retained
    p : float, optional
        The parameter :math:`p` of the negative binomial,
        :math:`1 / 1 + (ratio of the target database to the decoy database)`

    Returns
    -------
    float
        The theoretical number of incorrect target matches

    References
    ----------
    Levitsky, L. I., Ivanov, M. V., Lobas, A. A., & Gorshkov, M. V. (2017).
    Unbiased False Discovery Rate Estimation for Shotgun Proteomics Based
    on the Target-Decoy Approach. Journal of Proteome Research, 16(2), 393–397.
    https://doi.org/10.1021/acs.jproteome.6b00144
    """
    if t is None:
        return d + 1
    t = int(t)
    m = np.arange(t + 1, dtype=int)
    pi = np.exp(_log_pi(d, m, p))
    return ((m * pi).cumsum() / pi.cumsum())[t]


def expectation_correction(targets, decoys, ratio):
    """Estimate a correction for the number of decoys at a given
    score threshold for small data size.

    Parameters
    ----------
    targets : int
        The number of targets retained
    decoys : int
        The number of decoys retained
    ratio : float
        The ratio of target database to decoy database

    Returns
    -------
    float
        The number of decoys to add for the correction

    References
    ----------
    Levitsky, L. I., Ivanov, M. V., Lobas, A. A., & Gorshkov, M. V. (2017).
    Unbiased False Discovery Rate Estimation for Shotgun Proteomics Based
    on the Target-Decoy Approach. Journal of Proteome Research, 16(2), 393–397.
    https://doi.org/10.1021/acs.jproteome.6b00144
    """
    p = 1. / (1. + ratio)
    tfalse = _expectation(decoys, targets, p)
    return tfalse


class FDREstimatorBase(LoggingMixin):

    def summarize(self, name: Optional[str] = None):
        if name is None:
            name = "FDR"

        threshold_05, count_05 = self.get_count_for_fdr(0.05)
        self.log(f"5% {name} = {threshold_05:0.3f} ({count_05})")
        threshold_01, count_01 = self.get_count_for_fdr(0.01)
        self.log(f"1% {name} = {threshold_01:0.3f} ({count_01})")

    def score(self, spectrum_match: SpectrumMatch, assign: bool=False) -> float:
        raise NotImplementedError()

    def score_all(self, solution_set: SpectrumSolutionSet):
        raise NotImplementedError()

    def plot(self, ax=None):
        raise NotImplementedError()

    def get_count_for_fdr(self, threshold: float) -> Tuple[float, int]:
        raise NotImplementedError()


class TargetDecoyAnalyzer(FDREstimatorBase):
    """Estimate the False Discovery Rate using the Target-Decoy method.

    Attributes
    ----------
    database_ratio : float
        The ratio of the size of the target database to the decoy database
    target_weight : float
        A weight (less than 1.0) to put on target matches to make them weaker
        than decoys in situations where there is little data.
    decoy_correction : Number
        A quantity to use to correct for decoys, and if non-zero,
        will indicate that the negative binomial correction for decoys should be
        used.
    decoy_pseudocount : Number
        The value to report when querying the decoy count for a score exceeding
        the maximum score of a decoy match. This is distinct from `decoy_correction`
    decoy_count : int
        The total number of decoys
    decoys : list
        The decoy matches to consider
    n_decoys_at : dict
        The number of decoy matches above each threshold
    n_targets_at : dict
        The number of target matches above each threshold
    target_count : int
        The total number of targets
    targets : list
        The target matches to consider
    thresholds : list
        The distinct score thresholds
    with_pit : bool
        Whether or not to use the "percent incorrect target" adjustment
    """
    n_targets_at: NearestValueLookUp
    n_decoys_at: NearestValueLookUp

    targets: Sequence[SpectrumMatch]
    decoys: Sequence[SpectrumMatch]

    target_count: int
    decoy_count: int

    database_ratio: float
    decoy_correction: float

    target_weight: float
    decoy_pseudocount: float

    def __init__(self, target_series, decoy_series, with_pit=False, decoy_correction=0, database_ratio=1.0,
                 target_weight=1.0, decoy_pseudocount=1.0):
        self.targets = target_series
        self.decoys = decoy_series
        self.target_count = len(target_series)
        self.decoy_count = len(decoy_series)
        self.database_ratio = database_ratio
        self.target_weight = target_weight
        self.with_pit = with_pit
        self.decoy_correction = decoy_correction
        self.decoy_pseudocount = decoy_pseudocount

        self._calculate_thresholds()
        self._q_value_map = self.calculate_q_values()

    def get_score(self, spectrum_match: SpectrumMatch) -> float:
        return spectrum_match.score

    def has_score(self, spectrum_match: SpectrumMatch) -> bool:
        return hasattr(spectrum_match, 'score')

    def pack(self):
        self.targets = []
        self.decoys = []

    def _calculate_thresholds(self):
        self.n_targets_at = NearestValueLookUp([])
        self.n_decoys_at = NearestValueLookUp([])

        target_series = self.targets
        decoy_series = self.decoys

        if len(target_series) and self.has_score(target_series[0]):
            target_series = np.array([self.get_score(t)
                                      for t in target_series], dtype=float)
        else:
            target_series = np.array(target_series, dtype=float)

        if len(decoy_series) and self.has_score(decoy_series[0]):
            decoy_series = np.array([self.get_score(t)
                                     for t in decoy_series], dtype=float)
        else:
            decoy_series = np.array(decoy_series, dtype=float)

        thresholds = np.unique(
            np.sort(np.concatenate([target_series, decoy_series]))
        )

        self.thresholds = thresholds
        if len(thresholds) > 0:
            self.n_targets_at = ArrayScoreThresholdCounter(
                target_series, self.thresholds).counts_above_threshold
            self.n_decoys_at = ArrayScoreThresholdCounter(
                decoy_series, self.thresholds).counts_above_threshold
        else:
            self.n_targets_at = NearestValueLookUp([])
            self.n_decoys_at = NearestValueLookUp([])

    def n_decoys_above_threshold(self, threshold: float) -> int:
        try:
            if threshold > self.n_decoys_at.max_key():
                return self.decoy_pseudocount + self.decoy_correction
            return self.n_decoys_at[threshold] + self.decoy_correction
        except (IndexError, KeyError):
            if len(self.n_decoys_at) == 0:
                return self.decoy_correction
            else:
                raise

    def n_targets_above_threshold(self, threshold: float) -> int:
        try:
            return self.n_targets_at[threshold]
        except (IndexError, KeyError):
            if len(self.n_targets_at) == 0:
                return 0
            else:
                raise

    def expectation_correction(self, t: int, d: int) -> float:
        return expectation_correction(t, d, self.database_ratio)

    def target_decoy_ratio(self, cutoff: float) -> float:

        decoys_at = self.n_decoys_above_threshold(cutoff)
        targets_at = self.n_targets_above_threshold(cutoff)
        decoy_correction = 0
        if self.decoy_correction:
            try:
                decoy_correction = self.expectation_correction(
                    targets_at, decoys_at)
            except Exception as ex:
                print(ex)
        try:
            ratio = (decoys_at + decoy_correction) / float(
                targets_at * self.database_ratio * self.target_weight)
        except ZeroDivisionError:
            ratio = (decoys_at + decoy_correction)
        return ratio, targets_at, decoys_at

    def estimate_percent_incorrect_targets(self, cutoff: float) -> float:
        target_cut = self.target_count - self.n_targets_above_threshold(cutoff)
        decoy_cut = self.decoy_count - self.n_decoys_above_threshold(cutoff)
        percent_incorrect_targets = target_cut / float(decoy_cut)

        return percent_incorrect_targets

    def estimate_fdr(self, cutoff: float) -> float:
        if self.with_pit:
            percent_incorrect_targets = self.estimate_percent_incorrect_targets(
                cutoff)
        else:
            percent_incorrect_targets = 1.0
        return percent_incorrect_targets * self.target_decoy_ratio(cutoff)[0]

    def calculate_q_values(self):
        # Thresholds in ascending order
        thresholds = self.thresholds
        mapping = {}
        last_score = float('inf')
        last_q_value = 0
        for threshold in thresholds:
            try:
                q_value = self.estimate_fdr(threshold)
                # If a worse score has a lower q-value than a better score, use that q-value
                # instead.
                if last_q_value < q_value and last_score < threshold:
                    q_value = last_q_value
                last_q_value = q_value
                last_score = threshold
                mapping[threshold] = q_value
            except ZeroDivisionError:
                mapping[threshold] = 1.
        return NearestValueLookUp(mapping)

    def score_for_fdr(self, fdr_estimate: float) -> float:
        i = -1
        n = len(self.q_value_map)
        if n == 0:
            return 0
        for _score, fdr in self.q_value_map.items:
            i += 1
            if fdr_estimate >= fdr:
                if i < n:
                    cella = self.q_value_map.items[i]
                else:
                    cella = ScoreCell(fdr_estimate, 0)

                if i - 1 >= 0:
                    cellb = self.q_value_map.items[i - 1]
                else:
                    cellb = ScoreCell(fdr_estimate, 0)

                if i + 1 < n:
                    cellc = self.q_value_map.items[i + 1]
                else:
                    cellc = ScoreCell(fdr_estimate, 0)
                distance_a = abs(fdr_estimate - cella.value)
                distance_b = abs(fdr_estimate - cellb.value)
                distance_c = abs(fdr_estimate - cellc.value)
                min_distance = min(distance_a, distance_b, distance_c)
                if min_distance == distance_a:
                    return cella.score
                elif min_distance == distance_b:
                    return cellb.score
                else:
                    return cellc.score
        return float('inf')

    def get_count_for_fdr(self, fdr_estimate: float) -> Tuple[float, int]:
        threshold = self.score_for_fdr(fdr_estimate)
        count = self.n_targets_above_threshold(threshold)
        return threshold, count

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        thresholds = sorted(self.thresholds, reverse=False)
        target_counts = np.array(
            [self.n_targets_above_threshold(i) for i in thresholds])
        decoy_counts = np.array([self.n_decoys_above_threshold(i)
                                 for i in thresholds])
        fdr = np.array([self.q_value_map[i] for i in thresholds])
        try:
            at_5_percent = np.where(fdr < 0.05)[0][0]
        except IndexError:
            at_5_percent = -1
        try:
            at_1_percent = np.where(fdr < 0.01)[0][0]
        except IndexError:
            at_1_percent = -1
        line1 = ax.plot(thresholds, target_counts,
                        label='Target', color='steelblue')
        line2 = ax.plot(thresholds, decoy_counts, label='Decoy', color='coral')
        tline5 = ax.vlines(
            thresholds[at_5_percent], 0, np.max(target_counts), linestyle='--', color='green',
            lw=0.75, label='5% FDR')
        tline1 = ax.vlines(
            thresholds[at_1_percent], 0, np.max(target_counts), linestyle='--', color='skyblue',
            lw=0.75, label='1% FDR')
        ax.set_ylabel("# Matches Retained")
        ax.set_xlabel("Score")
        ax2 = ax.twinx()
        line3 = ax2.plot(thresholds, fdr, label='FDR',
                         color='grey', linestyle='--')
        ax2.set_ylabel("FDR")
        ax.legend([line1[0], line2[0], line3[0], tline5, tline1],
                  ['Target', 'Decoy', 'FDR', "5% FDR", "1% FDR"], frameon=False)

        lo, hi = ax.get_ylim()
        lo = max(lo, 0)
        ax.set_ylim(lo, hi)
        lo, hi = ax2.get_ylim()
        ax2.set_ylim(0, hi)

        lo, hi = ax.get_xlim()
        ax.set_xlim(-1, hi)
        lo, hi = ax2.get_xlim()
        ax2.set_xlim(-1, hi)
        return ax

    def q_values(self):
        q_map = self._q_value_map
        if len(q_map) == 0:
            import warnings
            warnings.warn("No FDR estimate what possible.")
            for target in self.targets:
                target.q_value = 0.0
            for decoy in self.decoys:
                decoy.q_value = 0.0
            return
        for target in self.targets:
            try:
                target.q_value = q_map[self.get_score(target)]
            except IndexError:
                target.q_value = 0.0
        for decoy in self.decoys:
            try:
                decoy.q_value = q_map[self.get_score(decoy)]
            except IndexError:
                decoy.q_value = 0.0

    def score(self, spectrum_match, assign=True):
        try:
            q_value = self._q_value_map[self.get_score(spectrum_match)]
        except IndexError:
            import warnings
            warnings.warn("Empty q-value mapping. q-value will be 0.")
            q_value = 0.0
        if assign:
            spectrum_match.q_value = q_value
        return q_value

    def score_all(self, solution_set):
        for spectrum_match in solution_set:
            self.score(spectrum_match, assign=True)
        solution_set.q_value = solution_set.best_solution().q_value
        return solution_set

    @property
    def q_value_map(self):
        return self._q_value_map

    @property
    def fdr_map(self):
        return self._q_value_map


class GroupwiseTargetDecoyAnalyzer(FDREstimatorBase):
    _grouping_labels = None
    targets: Sequence[SpectrumMatch]
    decoys: Sequence[SpectrumMatch]

    groups: List[List[Tuple[List[SpectrumMatch], List[SpectrumMatch]]]]
    group_fits: List[TargetDecoyAnalyzer]
    grouping_functions: List[Callable[[SpectrumMatch], bool]]
    grouping_labels: List[str]

    def __init__(self, target_series, decoy_series, with_pit=False, grouping_functions=None, decoy_correction=0,
                 database_ratio=1.0, target_weight=1.0, decoy_pseudocount=1.0, grouping_labels=None):
        if grouping_functions is None:
            grouping_functions = [lambda x: True]
        if grouping_labels is None:
            grouping_labels = ["Group %d" %
                               i for i in range(1, len(grouping_functions) + 1)]
        self.targets = target_series
        self.decoys = decoy_series
        self.with_pit = with_pit
        self.grouping_labels = grouping_labels
        self.grouping_functions = []
        self.groups = []
        self.group_fits = []
        self.decoy_pseudocount = decoy_pseudocount
        self.decoy_correction = decoy_correction
        self.database_ratio = database_ratio
        self.target_weight = target_weight

        for fn in grouping_functions:
            self.add_group(fn)

        self.partition()

    @property
    def grouping_labels(self):
        if self._grouping_labels is None:
            self._grouping_labels = [
                "Group %d" % i for i in range(1, len(self.grouping_functions) + 1)]
        return self._grouping_labels

    @grouping_labels.setter
    def grouping_labels(self, labels):
        self._grouping_labels = labels

    def pack(self):
        self.targets = []
        self.decoys = []
        self.groups = [[] for g in self.groups]
        for fit in self.group_fits:
            fit.pack()

    def partition(self):
        for target in self.targets:
            i = self.find_group(target)
            self.groups[i][0].append(target)
        for decoy in self.decoys:
            i = self.find_group(decoy)
            self.groups[i][1].append(decoy)
        for group in self.groups:
            fit = TargetDecoyAnalyzer(
                *group, with_pit=self.with_pit,
                decoy_correction=self.decoy_correction,
                database_ratio=self.database_ratio,
                target_weight=self.target_weight,
                decoy_pseudocount=self.decoy_pseudocount)
            self.group_fits.append(fit)

    def add_group(self, fn):
        self.grouping_functions.append(fn)
        self.groups.append(([], []))
        return len(self.groups)

    def find_group(self, spectrum_match):
        for i, fn in enumerate(self.grouping_functions):
            if fn(spectrum_match):
                return i
        return None

    def q_values(self):
        for group in self.group_fits:
            group.q_values()

    def score(self, spectrum_match, assign=True):
        i = self.find_group(spectrum_match)
        fit = self.group_fits[i]
        return fit.score(spectrum_match, assign=assign)

    def score_all(self, solution_set):
        for spectrum_match in solution_set:
            self.score(spectrum_match, assign=True)
        solution_set.q_value = solution_set.best_solution().q_value
        return solution_set

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        ax2 = ax.twinx()
        lines = []
        labels = []
        for _, (group_fit, label) in enumerate(zip(self.group_fits, self.grouping_labels)):
            thresholds = sorted(group_fit.thresholds, reverse=False)
            target_counts = np.array(
                [group_fit.n_targets_above_threshold(i) for i in thresholds])
            decoy_counts = np.array(
                [group_fit.n_decoys_above_threshold(i) for i in thresholds])

            fdr = np.array([group_fit.q_value_map[i] for i in thresholds])
            line1 = ax.plot(thresholds, target_counts,
                            label='%s Target' % label, )
            lines.append(line1[0])
            labels.append(line1[0].get_label())

            line2 = ax.plot(thresholds, decoy_counts,
                            label='%s Decoy' % label, )
            lines.append(line2[0])
            labels.append(line2[0].get_label())

            line3 = ax2.plot(thresholds, fdr, label='%s FDR' % label,
                             linestyle='--')
            lines.append(line3[0])
            labels.append(line3[0].get_label())

        ax.set_ylabel("# Matches Retained")
        ax.set_xlabel("Score")

        ax2.set_ylabel("FDR")
        ax.legend(lines, labels, frameon=False)

        lo, hi = ax.get_ylim()
        lo = max(lo, 0)
        ax.set_ylim(lo, hi)
        lo, hi = ax2.get_ylim()
        ax2.set_ylim(0, hi)

        lo, hi = ax.get_xlim()
        ax.set_xlim(-1, hi)
        lo, hi = ax2.get_xlim()
        ax2.set_xlim(-1, hi)
        return ax

    def summarize(self, name: Optional[str]=None):
        if name is None:
            name = "FDR"

        for group_name, group_fit in zip(self.grouping_labels, self.group_fits):
            group_fit.summarize(f"{group_name} {name}")


class PeptideScoreTargetDecoyAnalyzer(TargetDecoyAnalyzer):
    """
    A :class:`TargetDecoyAnalyzer` subclass for directly
    handling :class:`~.MultiScoreSpectrumMatch` instances.
    """

    def get_score(self, spectrum_match):
        return spectrum_match.score_set.peptide_score

    def has_score(self, spectrum_match):
        try:
            spectrum_match.score_set.peptide_score
            return True
        except AttributeError:
            return False



from typing import Dict, Optional
from sqlalchemy.orm.session import object_session

from sqlalchemy import (
    Column, Numeric, Integer, String, ForeignKey, PickleType,
    Boolean, select)
from sqlalchemy.orm import relationship, backref, deferred
from sqlalchemy.ext.mutable import MutableDict, MutableList

import numpy as np

from ms_peak_picker import FittedPeak as MemoryFittedPeak, PeakIndex, PeakSet
from ms_deisotope import DeconvolutedPeak as MemoryDeconvolutedPeak, DeconvolutedPeakSet
from ms_deisotope.peak_set import Envelope
from ms_deisotope.averagine import (
    mass_charge_ratio)

from ms_deisotope.data_source.common import (
    ProcessedScan, PrecursorInformation as MemoryPrecursorInformation, ChargeNotProvided)


from .base import (Base, Mass, HasUniqueName)
from .utils import chunkiter


class SampleRun(Base, HasUniqueName):
    __tablename__ = "SampleRun"

    id = Column(Integer, primary_key=True, autoincrement=True)

    ms_scans = relationship(
        "MSScan", backref=backref("sample_run"), lazy='dynamic')
    sample_type = Column(String(128))

    completed = Column(Boolean(), default=False, nullable=False)

    def __repr__(self):
        return "SampleRun(id=%d, name=%s)" % (self.id, self.name)


class MSScan(Base):
    __tablename__ = "MSScan"

    id = Column(Integer, primary_key=True, autoincrement=True)
    index = Column(Integer, index=True)
    ms_level = Column(Integer)
    scan_time = Column(Numeric(10, 5, asdecimal=False), index=True)
    title = Column(String(512))
    scan_id = Column(String(512), index=True)
    sample_run_id = Column(Integer, ForeignKey(
        SampleRun.id, ondelete='CASCADE'), index=True)

    peak_set = relationship("FittedPeak", backref="scan", lazy="dynamic")
    deconvoluted_peak_set = relationship(
        "DeconvolutedPeak", backref='scan', lazy='dynamic')

    info = deferred(Column(MutableDict.as_mutable(PickleType)))

    def __repr__(self):
        f = "{}({}, {}, {}, {}".format(
            self.__class__.__name__, self.scan_id, self.ms_level, self.scan_time,
            self.deconvoluted_peak_set.count())
        if self.ms_level > 1:
            f = "%s %s" % (f, self.precursor_information)
        f += ")"
        return f

    def convert(self, fitted=False, deconvoluted=True):
        precursor_information = self.precursor_information
        # if precursor_information is not None:
        #     precursor_information = precursor_information.convert()


        session = object_session(self)
        conn = session.connection()

        if fitted:
            q = conn.execute(select([FittedPeak.__table__]).where(
                FittedPeak.__table__.c.scan_id == self.id)).fetchall()

            peak_set_items = list(
                map(make_memory_fitted_peak, q))

            peak_set = PeakSet(peak_set_items)
            peak_set._index()
            peak_index = PeakIndex(np.array([], dtype=np.float64), np.array(
                [], dtype=np.float64), peak_set)
        else:
            peak_index = PeakIndex(np.array([], dtype=np.float64), np.array(
                [], dtype=np.float64), PeakSet([]))

        if deconvoluted:
            q = conn.execute(select([DeconvolutedPeak.__table__]).where(
                DeconvolutedPeak.__table__.c.scan_id == self.id)).fetchall()

            deconvoluted_peak_set_items = list(
                map(make_memory_deconvoluted_peak, q))

            deconvoluted_peak_set = DeconvolutedPeakSet(
                deconvoluted_peak_set_items)
            deconvoluted_peak_set.reindex()
        else:
            deconvoluted_peak_set = DeconvolutedPeakSet([])

        info = self.info or {}

        (scan_id, scan_title, scan_ms_level, scan_time, scan_index) = session.query(
            MSScan.scan_id, MSScan.title, MSScan.ms_level, MSScan.scan_time, MSScan.index).filter(
                MSScan.id == self.id).first()

        scan = ProcessedScan(
            scan_id, scan_title, precursor_information, int(scan_ms_level),
            float(scan_time), scan_index, peak_index, deconvoluted_peak_set,
            activation=info.get('activation'))
        return scan

    @classmethod
    def bulk_load(cls, session, ids, chunk_size: int=512, scan_cache: Optional[Dict]=None):
        if scan_cache is None:
            scan_cache = {}
        out = scan_cache

        for chunk in chunkiter(ids, chunk_size):
            id_slice = list(filter(lambda x: x not in scan_cache,
                                   set(chunk)))
            res = session.execute(cls.__table__.select().where(
                cls.id.in_(id_slice)))
            pinfos = PrecursorInformation.bulk_load_from_product_ids(
                session, id_slice, chunk_size=chunk_size)
            for self in res.fetchall():
                peak_index = None
                deconvoluted_peak_set = DeconvolutedPeakSet([])
                scan = ProcessedScan(
                    self.scan_id, self.scan_id, pinfos.get(self.id), int(self.ms_level),
                    float(self.scan_time), self.index, peak_index, deconvoluted_peak_set,
                    activation=self.info.get('activation'))
                out[self.id] = scan
        return [out[i] for i in ids]

    @classmethod
    def _serialize_scan(cls, scan, sample_run_id=None):
        db_scan = cls(
            index=scan.index, ms_level=scan.ms_level,
            scan_time=float(scan.scan_time), title=scan.title,
            scan_id=scan.id, sample_run_id=sample_run_id,
            info={'activation': scan.activation})
        return db_scan

    @classmethod
    def serialize(cls, scan, sample_run_id=None):
        db_scan = cls._serialize_scan(scan, sample_run_id)
        db_scan.peak_set = map(FittedPeak.serialize, scan.peak_set)
        db_scan.deconvoluted_peak_set = map(
            DeconvolutedPeak.serialize, scan.deconvoluted_peak_set)
        return db_scan

    @classmethod
    def serialize_bulk(cls, scan, sample_run_id, session, fitted=True, deconvoluted=True):
        db_scan = cls._serialize_scan(scan, sample_run_id)

        session.add(db_scan)
        session.flush()

        if fitted:
            FittedPeak._serialize_bulk_list(scan.peak_set, db_scan.id, session)
        if deconvoluted:
            DeconvolutedPeak._serialize_bulk_list(
                scan.deconvoluted_peak_set, db_scan.id, session)
        return db_scan


class PrecursorInformation(Base):
    __tablename__ = "PrecursorInformation"

    id = Column(Integer, primary_key=True)
    sample_run_id = Column(Integer, ForeignKey(
        SampleRun.id, ondelete='CASCADE'), index=True)

    precursor_id = Column(Integer, ForeignKey(
        MSScan.id, ondelete='CASCADE'), index=True)
    precursor = relationship(MSScan, backref=backref("product_information"),
                             primaryjoin="PrecursorInformation.precursor_id == MSScan.id",
                             uselist=False)
    product_id = Column(Integer, ForeignKey(
        MSScan.id, ondelete='CASCADE'), index=True)
    product = relationship(MSScan, backref=backref("precursor_information", uselist=False),
                           primaryjoin="PrecursorInformation.product_id == MSScan.id",
                           uselist=False)
    neutral_mass = Mass()
    charge = Column(Integer)
    intensity = Column(Numeric(16, 4, asdecimal=False))
    defaulted = Column(Boolean)
    orphan = Column(Boolean)

    @property
    def extracted_neutral_mass(self):
        return self.neutral_mass

    @property
    def extracted_charge(self):
        return self.charge

    @property
    def extracted_intensity(self):
        return self.intensity

    def __repr__(self):
        return "DBPrecursorInformation({}, {}, {})".format(
            self.precursor.scan_id, self.neutral_mass, self.charge)

    def convert(self, data_source=None):
        precursor_id = None
        if self.precursor is not None:
            precursor_id = self.precursor.scan_id
        product_id = None
        if self.product is not None:
            product_id = self.product.scan_id
        charge = self.charge
        if charge == 0:
            charge = ChargeNotProvided
        conv = MemoryPrecursorInformation(
            mass_charge_ratio(self.neutral_mass, charge), self.intensity, charge,
            precursor_id, data_source, self.neutral_mass, charge,
            self.intensity, self.defaulted, self.orphan, product_scan_id=product_id)
        return conv

    @classmethod
    def bulk_load_from_product_ids(cls, session, ids, chunk_size: int=512, data_source=None) -> Dict[int, MemoryPrecursorInformation]:
        out = {}

        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(cls.__table__.select().where(cls.product_id.in_(chunk)))
            for self in res.fetchall():
                precursor_id = None
                if self.precursor_id is not None:
                    precursor_id = session.execute(select([MSScan.scan_id]).where(
                        MSScan.id == self.precursor_id)).scalar()

                product_id = None
                if self.product_id is not None:
                    product_id = session.execute(select([MSScan.scan_id]).where(
                        MSScan.id == self.product_id)).scalar()

                charge = self.charge
                if charge == 0:
                    charge = ChargeNotProvided

                conv = MemoryPrecursorInformation(
                    mass_charge_ratio(self.neutral_mass, charge), self.intensity, charge,
                    precursor_id, data_source, self.neutral_mass, charge,
                    self.intensity, self.defaulted, self.orphan, product_scan_id=product_id)
                out[self.product_id] = conv
        return out

    @classmethod
    def serialize(cls, inst, precursor, product, sample_run_id):
        charge = inst.extracted_charge
        if charge == ChargeNotProvided:
            charge = 0
        db_pi = PrecursorInformation(
            precursor_id=precursor.id, product_id=product.id,
            charge=charge, intensity=inst.extracted_intensity,
            neutral_mass=inst.extracted_neutral_mass, sample_run_id=sample_run_id,
            defaulted=inst.defaulted, orphan=inst.orphan)
        return db_pi


class PeakMixin(object):
    mz = Mass(False)
    intensity = Column(Numeric(16, 4, asdecimal=False))
    full_width_at_half_max = Column(Numeric(7, 6, asdecimal=False))
    signal_to_noise = Column(Numeric(10, 3, asdecimal=False))
    area = Column(Numeric(12, 4, asdecimal=False))

    @classmethod
    def _serialize_bulk_list(cls, peaks, scan_id, session):
        out = cls._prepare_serialize_list(peaks, scan_id)
        session.bulk_save_objects(out)

    @classmethod
    def _prepare_serialize_list(cls, peaks, scan_id):
        out = []
        for peak in peaks:
            db_peak = cls.serialize(peak)
            db_peak.scan_id = scan_id
            out.append(db_peak)
        return out

    def __repr__(self):
        return "DB" + repr(self.convert())


def make_memory_fitted_peak(self):
    return MemoryFittedPeak(
        self.mz, self.intensity, self.signal_to_noise, -1, -1,
        self.full_width_at_half_max, (self.area if self.area is not None else 0.))


class FittedPeak(Base, PeakMixin):
    __tablename__ = "FittedPeak"

    id = Column(Integer, primary_key=True, autoincrement=True)
    scan_id = Column(Integer, ForeignKey(
        MSScan.id, ondelete='CASCADE'), index=True)

    convert = make_memory_fitted_peak

    @classmethod
    def serialize(cls, peak):
        return cls(
            mz=peak.mz, intensity=peak.intensity, signal_to_noise=peak.signal_to_noise,
            full_width_at_half_max=peak.full_width_at_half_max, area=peak.area)


def make_memory_deconvoluted_peak(self):
    return MemoryDeconvolutedPeak(
        self.neutral_mass, self.intensity, self.charge,
        self.signal_to_noise, -1, self.full_width_at_half_max,
        self.a_to_a2_ratio, self.most_abundant_mass, self.average_mass,
        self.score, Envelope(
            self.envelope), self.mz, None, self.chosen_for_msms,
        (self.area if self.area is not None else 0.))


class DeconvolutedPeak(Base, PeakMixin):
    __tablename__ = "DeconvolutedPeak"

    id = Column(Integer, primary_key=True, autoincrement=True)
    neutral_mass = Mass(False)
    average_mass = Mass(False)
    most_abundant_mass = Mass(False)

    charge = Column(Integer)

    score = Column(Numeric(12, 6, asdecimal=False))
    scan_id = Column(Integer, ForeignKey(
        MSScan.id, ondelete='CASCADE'), index=True)
    envelope = Column(MutableList.as_mutable(PickleType))
    a_to_a2_ratio = Column(Numeric(8, 7, asdecimal=False))
    chosen_for_msms = Column(Boolean)

    def __eq__(self, other):
        return (abs(self.neutral_mass - other.neutral_mass) < 1e-5) and (
            abs(self.intensity - other.intensity) < 1e-5)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.mz, self.intensity, self.charge))

    convert = make_memory_deconvoluted_peak

    @classmethod
    def serialize(cls, peak):
        return cls(
            mz=peak.mz, intensity=peak.intensity, signal_to_noise=peak.signal_to_noise,
            full_width_at_half_max=peak.full_width_at_half_max,
            neutral_mass=peak.neutral_mass, average_mass=peak.average_mass,
            most_abundant_mass=peak.most_abundant_mass, charge=peak.charge, score=peak.score,
            envelope=list(peak.envelope), a_to_a2_ratio=peak.a_to_a2_ratio,
            chosen_for_msms=peak.chosen_for_msms, area=(peak.area if peak.area is not None else 0.))


def serialize_scan_bunch(session, bunch, sample_run_id=None):
    precursor = bunch.precursor
    db_precursor = MSScan.serialize(precursor, sample_run_id=sample_run_id)
    session.add(db_precursor)
    db_products = [MSScan.serialize(p, sample_run_id=sample_run_id)
                   for p in bunch.products]
    session.add_all(db_products)
    session.flush()
    for scan, db_scan in zip(bunch.products, db_products):
        pi = scan.precursor_information
        db_pi = PrecursorInformation.serialize(
            pi, db_precursor, db_scan, sample_run_id)
        session.add(db_pi)
    session.flush()
    return db_precursor, db_products


def serialize_scan_bunch_bulk(session, bunch, sample_run_id, ms1_fitted=True, msn_fitted=True):
    precursor = bunch.precursor
    db_precursor = MSScan.serialize_bulk(
        precursor, sample_run_id, session, fitted=ms1_fitted)
    db_products = [MSScan.serialize_bulk(p, sample_run_id, session, fitted=msn_fitted)
                   for p in bunch.products]
    for scan, db_scan in zip(bunch.products, db_products):
        pi = scan.precursor_information
        db_pi = PrecursorInformation(
            precursor_id=db_precursor.id, product_id=db_scan.id,
            charge=pi.extracted_charge, intensity=pi.extracted_intensity,
            neutral_mass=pi.extracted_neutral_mass, sample_run_id=sample_run_id)
        session.add(db_pi)
    session.flush()
    return db_precursor, db_products



cimport cython
cimport numpy as np


@cython.boundscheck(False)
cpdef bint is_diagonal(np.ndarray[double, ndim=2] mat):
    cdef:
        size_t i, j, count
        size_t n, m
    n = mat.shape[0]
    m = mat.shape[1]
    if n != m:
        return False
    count = 0
    for i in range(n):
        for j in range(m):
            if i == j:
                continue
            count += mat[i, j] != 0
    return count == 0


# A replacement for a namedtuple
cdef class GlycanPriorRecord(object):

    def __init__(self, score, matched):
        self.score = score
        self.matched = matched

    def __reduce__(self):
        return self.__class__, (self.score, self.matched)

    def __getitem__(self, i):
        if i == 0:
            return self.score
        elif i == 1:
            return self.matched
        else:
            raise IndexError(i)

    def __len__(self):
        return 2

    def __repr__(self):
        return "{self.__class__.__name__}({self.score}, {self.matched})".format(self=self)


import numpy as np
import math

from glycresoft.structure import FragmentMatchMap
from glycopeptidepy.structure.fragment import IonSeries

from .base import PeptideSpectrumMatcherBase
from .simple_score import SimpleCoverageScorer


class LogIntensityScorer(SimpleCoverageScorer):
    def __init__(self, scan, sequence, mass_shift=None, *args, **kwargs):
        super(LogIntensityScorer, self).__init__(scan, sequence, mass_shift, *args, **kwargs)

    def _intensity_score(self, error_tolerance=2e-5, *args, **kwargs):
        total = 0
        target_ion_series = {
            IonSeries.b, IonSeries.y, IonSeries.c, IonSeries.z}
        for peak, fragment in self.solution_map:
            if fragment.series not in target_ion_series:
                continue
            total += np.log10(peak.intensity)
        return total

    def calculate_score(self, error_tolerance=2e-5, *args, **kwargs):
        self._score = self._intensity_score(error_tolerance, *args, **kwargs) * self._coverage_score()
        return self._score


class HyperscoreScorer(PeptideSpectrumMatcherBase):
    def __init__(self, scan, sequence, mass_shift=None):
        super(HyperscoreScorer, self).__init__(scan, sequence, mass_shift)
        self._score = None
        self.solution_map = FragmentMatchMap()

    def _calculate_hyperscore(self, *args, **kwargs):
        n_term_intensity = 0
        c_term_intensity = 0
        n_term = 0
        c_term = 0
        for peak, fragment in self.solution_map:
            if fragment.series in (IonSeries.b, IonSeries.c):
                n_term += 1
                n_term_intensity += peak.intensity
            elif fragment.series in (IonSeries.y, IonSeries.z):
                c_term += 1
                c_term_intensity += peak.intensity
        hyper = 0
        factors = [math.factorial(n_term), math.factorial(c_term),
                   n_term_intensity, c_term_intensity]
        for f in factors:
            if not f:
                continue
            hyper += math.log(f)

        return hyper

    def calculate_score(self, error_tolerance=2e-5, *args, **kwargs):
        self._score = self._calculate_hyperscore(
            error_tolerance, *args, **kwargs)
        return self._score


try:
    _has_c = True
    from glycresoft._c.tandem.tandem_scoring_helpers import _calculate_hyperscore
    HyperscoreScorer._calculate_hyperscore = _calculate_hyperscore
except ImportError:
    _has_c = False



from .utils import figax, figure


class ArtistBase(object):

    def __init__(self, ax=None):
        self.ax = ax

    def create_axes(self):
        self.ax = figax()

    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    def _repr_html_(self):
        if self.ax is None:
            return repr(self)
        fig = (self.ax.get_figure())
        return fig._repr_html_()



from .glycan_lcms.render import GlycanChromatogramReportCreator
from .glycopeptide_lcmsms.render import GlycopeptideDatabaseSearchReportCreator


__all__ = [
    "GlycanChromatogramReportCreator",
    "GlycopeptideDatabaseSearchReportCreator"
]



try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (
    Column, Numeric, String)
from sqlalchemy.orm import validates
from sqlalchemy.orm.session import object_session

Base = declarative_base()


def Mass(index=True):
    return Column(Numeric(14, 6, asdecimal=False), index=index)


def find_by_name(session, model_class, name):
    return session.query(model_class).filter(model_class.name == name).first()


def make_unique_name(session, model_class, name):
    marked_name = name
    i = 1
    while find_by_name(session, model_class, marked_name) is not None:
        marked_name = "%s (%d)" % (name, i)
        i += 1
    return marked_name


class HasUniqueName(object):
    name = Column(String(128), default=u"", unique=True)
    uuid = Column(String(64), index=True, unique=True)

    @classmethod
    def make_unique_name(cls, session, name):
        return make_unique_name(session, cls, name)

    @classmethod
    def find_by_name(cls, session, name):
        return find_by_name(session, cls, name)

    @validates("name")
    def ensure_unique_name(self, key, name):
        session = object_session(self)
        if session is not None:
            model_class = self.__class__
            name = make_unique_name(session, model_class, name)
            return name
        else:
            return name



'''Implements a multiprocessing deconvolution algorithm
'''

from ms_deisotope.tools.deisotoper.scan_generator import ScanGenerator as _ScanGenerator

from glycresoft.task import (
    TaskBase,
)

from glycresoft.config import get_configuration


user_config = get_configuration()
huge_tree = user_config.get("xml_huge_tree", False)


class ScanGenerator(TaskBase, _ScanGenerator):
    pass



from .glycan_source import (
    TextFileGlycanHypothesisSerializer, GlycanTransformer,
    TextFileGlycanCompositionLoader,
    GlycanCompositionHypothesisMerger,
    GlycanTypes,
    named_reductions,
    named_derivatizations)
from .constrained_combinatorics import (
    CombinatorialGlycanHypothesisSerializer, CombinatoricCompositionGenerator)
from .glycan_combinator import (
    GlycanCombinationSerializer, GlycanCombinationBuilder)
from .glyspace import (
    NGlycanGlyspaceHypothesisSerializer, OGlycanGlyspaceHypothesisSerializer,
    TaxonomyFilter)
from .synthesis import (
    SynthesisGlycanHypothesisSerializer, ExistingGraphGlycanHypothesisSerializer,
    GlycanCompositionEnzymeGraph, synthesis_register)
from .convert_analysis import (
    GlycanAnalysisHypothesisSerializer,
    GlycopeptideAnalysisGlycanCompositionExtractionHypothesisSerializer)
